import datetime as dti
import difflib
import hashlib
import json
import pathlib
import platform
import re
import subprocess  # nosec B404
import uuid
from typing import Any, Callable, Generator, Union, no_type_check

import yaml

import foran.foran as api  # type: ignore
from foran.report import generate_report  # type: ignore
from taksonomia.taksonomia import Taxonomy  # type: ignore

from liitos import (
    CONTEXT,
    ENCODING,
    KEYS_REQUIRED,
    LATEX_PAYLOAD_NAME,
    TOOL_VERSION_COMMAND_MAP,
    ToolKey,
    log,
)

PathLike = Union[str, pathlib.Path]

DOC_BASE = pathlib.Path('..', '..')
STRUCTURE_PATH = DOC_BASE / 'structure.yml'
IMAGES_FOLDER = 'images/'
DIAGRAMS_FOLDER = 'diagrams/'
PATCH_SPEC_NAME = 'patch.yml'
CHUNK_SIZE = 2 << 15
TS_FORMAT = '%Y-%m-%d %H:%M:%S.%f +00:00'
LOG_SEPARATOR = '- ' * 80
INTER_PROCESS_SYNC_SECS = 0.1
INTER_PROCESS_SYNC_ATTEMPTS = 10

IS_BORING = re.compile(r'\(.*texmf-dist/tex.*\.')


def hash_file(path: pathlib.Path, hasher: Union[Callable[..., Any], None] = None) -> str:
    """Return the SHA512 hex digest of the data from file."""
    if hasher is None:
        hasher = hashlib.sha512
    the_hash = hasher()
    with open(path, 'rb') as handle:
        while chunk := handle.read(CHUNK_SIZE):
            the_hash.update(chunk)
    return the_hash.hexdigest()


@no_type_check
def log_subprocess_output(pipe, prefix: str):
    for line in iter(pipe.readline, b''):  # b'\n'-separated lines
        cand = line.decode(encoding=ENCODING).rstrip()
        if IS_BORING.search(cand):
            log.debug(cand)
            continue
        if cand.strip().strip('[])yex'):
            if any(
                [
                    'microtype' in cand,
                    'xassoccnt' in cand,
                    'texlive/2022/texmf-dist/tex/' in cand,
                    cand == 'erns.sty)',
                    cand == '(see the transcript file for additional information)',
                    cand.startswith(r'Overfull \hbox ')
                    and cand.endswith(r'pt too wide) has occurred while \output is active'),
                ]
            ):
                log.debug(f'{prefix}: %s', cand)
            else:
                log.info(f'{prefix}: %s', cand)


@no_type_check
def vcs_probe():
    """Are we in front, on par, or behind with the upstream?"""
    CONTEXT['source_hash'] = 'info:plain:built-outside-of-version-control'
    CONTEXT['source_hint'] = 'info:plain:built-outside-of-version-control'
    try:
        repo = api.Repo('.', search_parent_directories=True)
        status = api.Status(repo)
        api.local_commits(repo, status)
        api.local_staged(repo, status)
        api.local_files(repo, status)
        CONTEXT['source_hash'] = f'sha1:{status.commit}'
        try:
            repo_root_folder = repo.git.rev_parse(show_toplevel=True)
            path = pathlib.Path(repo_root_folder)
            anchor = path.parent.name
            here = path.name
            CONTEXT['source_hint'] = f'{anchor}/{here}'
            yield f'Root     ({repo_root_folder})'
        except Exception:  # noqa
            yield 'WARNING - ignored exception when assessing repo root folder location'
        for line in generate_report(status):
            yield line.rstrip()
    except Exception:  # noqa
        yield 'WARNING - we seem to not be within a git repository clone'


def node_id() -> str:
    """Generate the build node identifier."""
    return str(uuid.uuid3(uuid.NAMESPACE_DNS, platform.node()))


def report_taxonomy(target_path: pathlib.Path) -> None:
    """Convenience function to report date, size, and checksums of the deliverable."""
    taxonomy = Taxonomy(target_path, excludes='', key_function='md5')
    for path in sorted(target_path.parent.rglob('*')):
        taxonomy.add_branch(path) if path.is_dir() else taxonomy.add_leaf(path)
    log.info('- Writing render/pdf folder taxonomy to inventory.json ...')
    taxonomy.dump(sink='inventory', format_type='json', base64_encode=False)

    stat = target_path.stat()
    size_bytes = stat.st_size
    mod_time = dti.datetime.fromtimestamp(stat.st_ctime, tz=dti.timezone.utc).strftime(TS_FORMAT)
    sha612_hash = hash_file(target_path, hashlib.sha512)
    sha256_hash = hash_file(target_path, hashlib.sha256)
    sha1_hash = hash_file(target_path, hashlib.sha1)
    md5_hash = hash_file(target_path, hashlib.md5)
    log.info('- Ephemeral:')
    log.info(f'  + name: {target_path.name}')
    log.info(f'  + size: {size_bytes} bytes')
    log.info(f'  + date: {mod_time}')
    log.info('- Characteristic:')
    log.info('  + Checksums:')
    log.info(f'    sha512:{sha612_hash}')
    log.info(f'    sha256:{sha256_hash}')
    log.info(f'      sha1:{sha1_hash}')
    log.info(f'       md5:{md5_hash}')
    log.info('  + Fonts:')


@no_type_check
def unified_diff(left: list[str], right: list[str], left_label: str = 'before', right_label: str = 'after'):
    """Derive the unified diff between left and right lists of strings as generator of strings."""
    for line in difflib.unified_diff(left, right, fromfile=left_label, tofile=right_label):
        yield line.rstrip()


@no_type_check
def log_unified_diff(left: list[str], right: list[str], left_label: str = 'before', right_label: str = 'after'):
    """Do the log bridging of the diff."""
    log.info(LOG_SEPARATOR)
    for line in unified_diff(left, right, left_label, right_label):
        for fine in line.split('\n'):
            log.info(fine)
    log.info(LOG_SEPARATOR)


@no_type_check
def ensure_separate_log_lines(sourcer: Callable, *args: Union[list[object], None]):
    """Wrapping idiom breaking up any strings containing newlines."""
    log.info(LOG_SEPARATOR)
    for line in sourcer(*args) if args else sourcer():
        for fine in line.split('\n'):
            log.info(fine)
    log.info(LOG_SEPARATOR)


@no_type_check
def delegate(command: list[str], marker: str, do_shell: bool = False) -> int:
    """Execute command in subprocess and follow requests."""
    try:
        process = subprocess.Popen(
            command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=do_shell  # nosec B602
        )
        with process.stdout:
            log_subprocess_output(process.stdout, marker)
        code = process.wait()
        if code < 0:
            log.error(f'{marker} process ({command}) was terminated by signal {-code}')
        elif code > 0:
            log.error(f'{marker} process ({command}) returned {code}')
        else:
            log.info(f'{marker} process succeeded')
    except Exception as err:
        log.error(f'failed executing tool with error: {err}')
        code = 42

    return code


@no_type_check
def report(on: ToolKey) -> int:
    """Execute the tool specific version command."""
    tool_context = TOOL_VERSION_COMMAND_MAP.get(on, {})
    tool_version_call_text = str(tool_context.get('command', '')).strip()
    tool_version_call = tool_version_call_text.split()
    tool_reason_banner = str(tool_context.get('banner', 'No reason for the tool known')).strip()
    if not tool_version_call:
        log.warning(f'cowardly avoiding undefined call for tool key ({on})')
        log.info(f'- known tool keys are: ({", ".join(sorted(TOOL_VERSION_COMMAND_MAP))})')
        return 42

    log.info(LOG_SEPARATOR)
    log.info(f'requesting tool version information from environment per ({tool_version_call})')
    log.info(f'- {tool_reason_banner}')
    code = delegate(tool_version_call, f'tool-version-of-{on}')
    log.info(LOG_SEPARATOR)

    return code


@no_type_check
def execute_filter(
    the_filter: Callable,
    head: str,
    backup: str,
    label: str,
    text_lines: list[str],
    lookup: Union[dict[str, str], None] = None,
) -> list[str]:
    """Chain filter calls by storing in and out lies in files and return the resulting lines."""
    log.info(LOG_SEPARATOR)
    log.info(head)
    doc_before_caps_patch = backup
    with open(doc_before_caps_patch, 'wt', encoding=ENCODING) as handle:
        handle.write('\n'.join(text_lines))
    patched_lines = the_filter(text_lines, lookup=lookup)
    with open(LATEX_PAYLOAD_NAME, 'wt', encoding=ENCODING) as handle:
        handle.write('\n'.join(patched_lines))
    log.info(f'diff of the ({label}) filter result:')
    log_unified_diff(text_lines, patched_lines)

    return patched_lines


@no_type_check
def load_target(
    target_code: str, facet_code: str, structure_path: PathLike = STRUCTURE_PATH
) -> tuple[bool, dict[str, str]]:
    """DRY."""
    if not structure_path.is_file() or not structure_path.stat().st_size:
        log.error(f'render failed to find non-empty structure file at {structure_path}')
        return False, {}

    with open(structure_path, 'rt', encoding=ENCODING) as handle:
        structure = yaml.safe_load(handle)

    targets = sorted(structure.keys())

    if not targets:
        log.error(f'structure at ({structure_path}) does not provide any targets')
        return False, {}

    if target_code not in targets:
        log.error(f'structure does not provide ({target_code})')
        return False, {}

    if len(targets) != 1:
        log.warning(f'unexpected count of targets ({len(targets)}) from ({targets})')
        return True, {}

    target = targets[0]
    facets = sorted(list(facet.keys())[0] for facet in structure[target])
    log.info(f'found single target ({target}) with facets ({facets})')

    if facet_code not in facets:
        log.error(f'structure does not provide facet ({facet_code}) for target ({target_code})')
        return False, {}

    aspect_map = {}
    for data in structure[target]:
        if facet_code in data:
            aspect_map = data[facet_code]
            break
    missing_keys = [key for key in KEYS_REQUIRED if key not in aspect_map]
    if missing_keys:
        log.error(
            f'structure does not provide all expected aspects {sorted(KEYS_REQUIRED)}'
            f' for target ({target_code}) and facet ({facet_code})'
        )
        log.error(f'- the found aspects: {sorted(aspect_map.keys())}')
        log.error(f'- missing aspects:   {sorted(missing_keys)}')
        return False, {}

    if sorted(aspect_map.keys()) != sorted(KEYS_REQUIRED):
        log.debug(
            f'structure does not strictly provide the expected aspects {sorted(KEYS_REQUIRED)}'
            f' for target ({target_code}) and facet ({facet_code})'
        )
        log.debug(f'- found the following aspects instead:                   {sorted(aspect_map.keys())} instead')

    return True, aspect_map


@no_type_check
def mermaid_captions_from_json_ast(json_ast_path: Union[str, pathlib.Path]) -> dict[str, str]:
    """Separation of concerns."""
    doc = json.load(open(json_ast_path, 'rt', encoding=ENCODING))
    blocks = doc['blocks']
    mermaid_caption_map = {}
    for b in blocks:
        if b['t'] == 'CodeBlock' and b['c'][0]:
            try:
                is_mermaid = b['c'][0][1][0] == 'mermaid'
                atts = b['c'][0][2]
            except IndexError:
                continue

            if not is_mermaid:
                continue
            m_caption, m_filename, m_format, m_loc = '', '', '', ''
            for k, v in atts:
                if k == 'caption':
                    m_caption = v
                elif k == 'filename':
                    m_filename = v
                elif k == 'format':
                    m_format = v
                elif k == 'loc':
                    m_loc = v
                else:
                    pass
            token = f'{m_loc}/{m_filename}.{m_format}'  # noqa
            if token in mermaid_caption_map:
                log.warning('Duplicate token, same caption?')
                log.warning(f'-   prior: {token} -> {m_caption}')
                log.warning(f'- current: {token} -> {mermaid_caption_map[token]}')
            mermaid_caption_map[token] = m_caption
    return mermaid_caption_map


def remove_target_region_gen(text_lines: list[str], from_cut: str, thru_cut: str) -> Generator[str, None, None]:
    """Return generator that yields only the lines beyond the cut mark region skipping lines in [from, thru]."""
    in_section = False
    for line in text_lines:
        if not in_section:
            if from_cut in line:
                in_section = True
                continue
        if in_section:
            if thru_cut in line:
                in_section = False
            continue
        yield line



from collections.abc import Iterable
from typing import Union

from liitos import log

NO_LABEL = 'no-label-found-ERROR'


def is_include_graphics(text: str) -> bool:
    """Only DRY."""
    ntoken = r'\pandocbounded'  # nosec B105
    pos_after = len(ntoken)
    if len(text) > len(ntoken) and text.startswith(ntoken) and text[pos_after] in ('[', '{'):
        return True
    token = r'\includegraphics'  # nosec B105
    pos_after = len(token)
    return len(text) > len(token) and text.startswith(token) and text[pos_after] in ('[', '{')


def extract_image_path(include_graphics_line: str) -> str:
    """We had a bug, so we isolate in a function."""
    if include_graphics_line and 'pandocbounded{' in include_graphics_line:
        return include_graphics_line.split('{', 2)[2].rstrip().rstrip('}')
    if include_graphics_line and '{' in include_graphics_line:
        return include_graphics_line.split('{', 1)[1].rstrip().rstrip('}')
    else:
        return 'IMAGE_PATH_NOT_FOUND'


def inject(incoming: Iterable[str], lookup: Union[dict[str, str], None] = None) -> list[str]:
    """Later alligator."""
    outgoing = []
    modus = 'copy'
    label = NO_LABEL
    figure: list[str] = []
    caption: list[str] = []
    precondition = r'\begin{figure}'
    precondition_met = False
    for slot, line in enumerate(incoming):
        if line.startswith(precondition) and not precondition_met:
            log.info(f'start of a figure environment at line #{slot + 1}')
            precondition_met = True
            outgoing.append(line)
            continue

        if modus == 'copy':
            if is_include_graphics(line) and not precondition_met:
                log.warning(f'graphics include outside of a figure environment at line #{slot + 1}')
                log.debug(f'line#{slot + 1}|{line.rstrip()}')
                log.info('trying to fix temporarily ... watch for marker MISSING-CAPTION-IN-MARKDOWN')
                adhoc_label = 'FIX-AT-SOURCE'
                try:
                    lab = line.split('{', 1)[1]
                    lab = lab.rsplit('.', 1)[0]
                    lab = lab.rsplit('/', 1)[1]
                    adhoc_label = r'\label{fig:' + lab + '}'
                    log.info(adhoc_label)
                except Exception as err:
                    log.error(f'failed to extract generic label from {line.strip()} with err: {err}')
                captain = 'MISSING-CAPTION-IN-MARKDOWN'
                try:
                    token = extract_image_path(line)
                    log.debug(f'- looking up key ({token}) for captions ...')
                    if lookup is not None:
                        cand = lookup.get(token, None)
                        if cand is not None:
                            captain = cand
                            log.debug(f'  + found ({captain})')
                        else:
                            log.debug(f'  ? no match for key ({token})')
                    else:
                        log.debug('  + no lut?')
                except Exception as err:
                    log.error(
                        f'failed to extract file path token for caption lookup from {line.strip()} with err: {err}'
                    )
                outgoing.append(r'\begin{figure}' + '\n')
                outgoing.append(r'\centering' + '\n')
                outgoing.append(line)
                outgoing.append(r'\caption{' + captain + ' ' + adhoc_label + '}' + '\n')
                outgoing.append(r'\end{figure}' + '\n')
            elif is_include_graphics(line) and precondition_met:
                log.info(f'within a figure environment at line #{slot + 1}')
                log.info(line.rstrip())
                modus = 'figure'
                figure = [line]
                try:
                    lab = line.split('{', 1)[1]
                    lab = lab.rsplit('.', 1)[0]
                    lab = lab.rsplit('/', 1)[1]
                    label = r'\label{fig:' + lab + '}'
                    log.info(label)
                except Exception as err:
                    log.error(f'failed to extract generic label from {line.strip()} with err: {err}')
                caption = []
            else:
                log.debug(f'- copying {slot + 1 :3d}|{line.rstrip()}')
                outgoing.append(line)
            continue

        if modus == 'figure':
            if line.startswith(r'\caption{'):
                log.info(f'- found the caption start at line #{slot + 1}')
                caption.append(line)
                if not line.strip().endswith('}'):
                    log.info(f'- multi line caption at line #{slot + 1}')
                    modus = 'caption'
            elif line.startswith(r'\end{figure}'):
                log.info(f'end of figure env detected at line #{slot + 1}')
                outgoing.extend(figure)
                caption_text = '\n'.join(caption)
                if r'\label{' not in caption_text:
                    caption_text = f"{caption_text.rstrip().rstrip('}')} {label}" + '}'
                if caption_text == caption_text.rstrip():
                    caption_text += '\n'
                outgoing.append(caption_text)
                outgoing.append(line)
                modus = 'copy'
                label = NO_LABEL
                precondition_met = False
            else:
                log.debug(f'- figure continues at line #{slot + 1}')
                figure.append(line)
            continue

        if modus == 'caption':
            caption.append(line)
            if line.strip().endswith(r'}'):
                log.info(f'- caption read at line #{slot + 1}')
                modus = 'figure'
            continue

    return outgoing



import os
import pathlib
import tempfile

import liitos.placeholder as placeholder
from liitos import ENCODING

ALIEN_PATH = pathlib.Path('test', 'fixtures', 'template', 'alien')
EXTERNAL_PATH = pathlib.Path('test', 'fixtures', 'placeholder', 'this-resource-is-missing.svg')


def test_load_svg_placeholder():
    kind, data = placeholder.load_resource('placeholders/this-resource-is-missing.svg', False)
    assert kind == 'str'
    assert isinstance(data, str)


def test_load_svg_placeholder_external():
    kind, data = placeholder.load_resource(EXTERNAL_PATH, True)
    assert kind == 'str'
    assert isinstance(data, str)


def test_load_alien():
    kind, data = placeholder.load_resource(ALIEN_PATH, True)
    assert kind == 'bytes'
    assert isinstance(data, bytes)
    assert int(data.decode(encoding=ENCODING).strip()) == 42


def test_eject():
    with tempfile.TemporaryDirectory() as tmpdirname:
        assert placeholder.eject([tmpdirname]) == 0


def test_eject_default():
    prior = os.getcwd()
    with tempfile.TemporaryDirectory() as tmpdirname:
        os.chdir(tmpdirname)
        assert placeholder.eject() == 0
    os.chdir(prior)


def test_dump_placeholder_text():
    base = 'foo'
    suffix = '.svg'
    target_name = f'{base}{suffix}'
    with tempfile.TemporaryDirectory() as tmpdirname:
        code, msg = placeholder.dump_placeholder(pathlib.Path(tmpdirname) / target_name)
    assert code == 0
    assert msg.startswith(f'wrote text mode placeholder matching suffix ({suffix}) derived from target(')
    assert msg.endswith(f'/{target_name})')


def test_dump_placeholder_binary():
    base = 'foo'
    suffix = '.png'
    target_name = f'{base}{suffix}'
    with tempfile.TemporaryDirectory() as tmpdirname:
        code, msg = placeholder.dump_placeholder(pathlib.Path(tmpdirname) / target_name)
    assert code == 0
    assert msg.startswith(f'wrote binary mode placeholder matching suffix ({suffix}) derived from target(')
    assert msg.endswith(f'/{target_name})')


def test_dump_placeholder_miss():
    base = 'foo'
    suffix = '.does-not-exist'
    target_name = f'{base}{suffix}'
    with tempfile.TemporaryDirectory() as tmpdirname:
        code, msg = placeholder.dump_placeholder(pathlib.Path(tmpdirname) / target_name)
    assert code == 1
    assert msg.startswith(f'no placeholder found matching suffix ({suffix}) derived from target(')
    assert msg.endswith(f'/{target_name})')



import json
import liitos.labels as labels


def test_inject_on_empty_ok():
    assert labels.inject([]) == []


def test_inject_no_match_passes_through_ok():
    pass_me_through = ['a', 'b', 'caption', 'figure', 'table', 'label']
    assert labels.inject(pass_me_through) == pass_me_through


def test_inject_precondition():
    partial_matches = ['a', r'\begin{figure}', 'z']
    assert labels.inject(partial_matches) == partial_matches


def test_inject_not_precondition():
    partial_matches = ['a\n', r'\includegraphics{' + '\n', 'z\n']
    augmented_content = [
        'a\n',
        r'\begin{figure}' + '\n',
        r'\centering' + '\n',
        r'\includegraphics{' + '\n',
        r'\caption{MISSING-CAPTION-IN-MARKDOWN FIX-AT-SOURCE}' + '\n',
        r'\end{figure}' + '\n',
        'z\n',
    ]
    assert labels.inject(partial_matches) == augmented_content


def test_inject_precondition_and_include():
    correct_match = [
        'a\n',
        '\n',
        r'\begin{figure}' + '\n',
        r'\centering' + '\n',
        r'\includegraphics{images/blue.png}' + '\n',
        r'\caption{Caption Text Blue}' + '\n',
        r'\end{figure}' + '\n',
        '\n',
        'z\n',
    ]
    injected = [
        'a\n',
        '\n',
        r'\begin{figure}' + '\n',
        r'\centering' + '\n',
        r'\includegraphics{images/blue.png}' + '\n',
        r'\caption{Caption Text Blue \label{fig:blue}}' + '\n',
        r'\end{figure}' + '\n',
        '\n',
        'z\n',
    ]
    assert labels.inject(correct_match) == injected


def test_is_include_graphics():
    assert labels.is_include_graphics(r'\includegraphics[') is True
    assert labels.is_include_graphics(r'\includegraphics{') is True
    assert labels.is_include_graphics(r'\includegraphics[]') is True
    assert labels.is_include_graphics(r'\includegraphics{}\n') is True
    assert labels.is_include_graphics(r'\includegraphics') is False
    assert labels.is_include_graphics(r'\includegraphic[') is False
    assert labels.is_include_graphics('') is False


def test_extract_image_path():
    assert labels.extract_image_path(r'\includegraphics{a/b}') == 'a/b'
    assert labels.extract_image_path(r'\includegraphics[]{a/b}') == 'a/b'
    assert labels.extract_image_path(r'\includegraphics[]{}\n') == r'}\n'
    assert labels.extract_image_path('') == 'IMAGE_PATH_NOT_FOUND'


def test_inject_regression_bug48():
    lookup = json.load(open('test/fixtures/bugs/labels/lookup.json', 'rt', encoding='utf-8'))
    incoming = open('test/fixtures/bugs/labels/document-pre-labels.tex', 'rt', encoding='utf-8').readlines()
    reference = open('test/fixtures/bugs/labels/document-post-labels.tex', 'rt', encoding='utf-8').readlines()
    outgoing = labels.inject(incoming, lookup)
    assert outgoing == reference



"""Command line interface for splice (Finnish liitos) contributions."""

import copy
import datetime as dti
import logging
import os
import pathlib
import sys
from typing import Union

import typer

import liitos.approvals as sig
import liitos.changes as chg
import liitos.concat as cat
import liitos.eject as eje
import liitos.gather as gat
import liitos.meta as met
import liitos.render as ren
import liitos.tools as too
from liitos import (
    APP_NAME,
    APP_VERSION,
    APPROVALS_STRATEGY,
    DEFAULT_STRUCTURE_NAME,
    EXTERNALS,
    FILTER_CS_LIST,
    FROM_FORMAT_SPEC,
    KNOWN_APPROVALS_STRATEGIES,
    LOG_SEPARATOR,
    QUIET,
    TOOL_VERSION_COMMAND_MAP,
    TS_FORMAT_PAYLOADS,
    log,
)

app = typer.Typer(
    add_completion=False,
    context_settings={'help_option_names': ['-h', '--help']},
    no_args_is_help=True,
)

DocumentRoot = typer.Option(
    '',
    '-d',
    '--document-root',
    help='Root of the document tree to visit. Optional\n(default: positional tree root value)',
)
StructureName = typer.Option(
    DEFAULT_STRUCTURE_NAME,
    '-s',
    '--structure',
    help='structure mapping file (default: {gat.DEFAULT_STRUCTURE_NAME})',
)
TargetName = typer.Option(
    '',
    '-t',
    '--target',
    help='target document key',
)
FacetName = typer.Option(
    '',
    '-f',
    '--facet',
    help='facet key of target document',
)
FromFormatSpec = typer.Option(
    FROM_FORMAT_SPEC,
    '--from-format-spec',
    help='from format specification handed over to pandoc',
)
FilterCSList = typer.Option(
    'DEFAULT_FILTER',
    '-F',
    '--filters',
    help='comma separated list of filters handed over to pandoc (in order) or empty to apply no filter',
)
Verbosity = typer.Option(
    False,
    '-v',
    '--verbose',
    help='Verbose output (default is False)',
)
Strictness = typer.Option(
    False,
    '--strict',
    help='Ouput noisy warnings on console (default is False)',
)
OutputPath = typer.Option(
    '',
    '-o',
    '--output-path',
    help='Path to output unambiguous content to - like when ejecting a template',
)
LabelCall = typer.Option(
    '',
    '-l',
    '--label',
    help='optional label call to execute',
)
PatchTables = typer.Option(
    False,
    '-p',
    '--patch-tables',
    help='Patch tables EXPERIMENTAL (default is False)',
)
ApprovalsStrategy = typer.Option(
    '',
    '-a',
    '--approvals-strategy',
    help=f'optional approvals layout strategy in ({", ".join(KNOWN_APPROVALS_STRATEGIES)})',
)


@app.callback(invoke_without_command=True)
def callback(
    version: bool = typer.Option(
        False,
        '-V',
        '--version',
        help='Display the application version and exit',
        is_eager=True,
    )
) -> None:
    """
    Splice (Finnish liitos) contributions.
    """
    if version:
        typer.echo(f'{APP_NAME} version {APP_VERSION}')
        raise typer.Exit()


def _verify_call_vector(
    doc_root: str,
    doc_root_pos: str,
    verbose: bool,
    strict: bool,
    label: str = '',
    patch_tables: bool = False,
    from_format_spec: str = FROM_FORMAT_SPEC,
    filter_cs_list: str = '',
    approvals_strategy: str = '',
) -> tuple[int, str, str, dict[str, Union[bool, str]]]:
    """DRY"""
    log.debug(f'verifier received: {locals()}')
    doc = doc_root.strip()
    if not doc and doc_root_pos:
        doc = doc_root_pos
    if not doc:
        print('Document tree root required', file=sys.stderr)
        return 2, 'Document tree root required', '', {}

    doc_root_path = pathlib.Path(doc)
    if doc_root_path.exists():
        if not doc_root_path.is_dir():
            print(f'requested tree root at ({doc}) is not a folder reachable from ({os.getcwd()})', file=sys.stderr)
            return 2, f'requested tree root at ({doc}) is not a folder reachable from ({os.getcwd()})', '', {}
    else:
        print(f'requested tree root at ({doc}) does not exist as seen from ({os.getcwd()})', file=sys.stderr)
        return 2, f'requested tree root at ({doc}) does not exist as seen from ({os.getcwd()})', '', {}

    if not approvals_strategy:
        approvals_strategy = APPROVALS_STRATEGY
        log.info(f'Using value from environment for approvals strategy (APPROVALS_STRATEGY) == ({approvals_strategy})')
        if not approvals_strategy:
            approvals_strategy = KNOWN_APPROVALS_STRATEGIES[0]
            log.info(
                'No preference in environment for approvals strategy (APPROVALS_STRATEGY)'
                f' using default ({approvals_strategy})'
            )

    if approvals_strategy not in KNOWN_APPROVALS_STRATEGIES:
        approvals_strategy = KNOWN_APPROVALS_STRATEGIES[0]
        log.info(
            'Value in environment for approvals strategy (APPROVALS_STRATEGY)'
            f' not in ({", ".join(KNOWN_APPROVALS_STRATEGIES)}) - using default ({approvals_strategy})'
        )

    options: dict[str, Union[bool, str]] = {
        'quiet': QUIET and not verbose and not strict,
        'strict': strict,
        'verbose': verbose,
        'label': label,
        'patch_tables': patch_tables,
        'from_format_spec': from_format_spec if from_format_spec else FROM_FORMAT_SPEC,
        'filter_cs_list': filter_cs_list if filter_cs_list != 'DEFAULT_FILTER' else FILTER_CS_LIST,
        'approvals_strategy': approvals_strategy,
    }
    log.debug(f'Post verifier: {options=}')
    if verbose:
        logging.getLogger().setLevel(logging.DEBUG)
    elif options.get('quiet'):
        logging.getLogger().setLevel(logging.ERROR)
    return 0, '', doc, options


@app.command('verify')
def verify(  # noqa
    doc_root_pos: str = typer.Argument(''),
    doc_root: str = DocumentRoot,
    structure: str = StructureName,
    target: str = TargetName,
    facet: str = FacetName,
    verbose: bool = Verbosity,
    strict: bool = Strictness,
) -> int:
    """
    Verify the structure definition against the file system.
    """
    code, message, doc, options = _verify_call_vector(
        doc_root=doc_root, doc_root_pos=doc_root_pos, verbose=verbose, strict=strict
    )
    if code:
        log.error(message)
        return code

    return sys.exit(
        gat.verify(doc_root=doc, structure_name=structure, target_key=target, facet_key=facet, options=options)
    )


@app.command('approvals')
def approvals(  # noqa
    doc_root_pos: str = typer.Argument(''),
    doc_root: str = DocumentRoot,
    structure: str = StructureName,
    target: str = TargetName,
    facet: str = FacetName,
    verbose: bool = Verbosity,
    strict: bool = Strictness,
) -> int:
    """
    Weave in the approvals for facet of target within document root.
    """
    code, message, doc, options = _verify_call_vector(
        doc_root=doc_root, doc_root_pos=doc_root_pos, verbose=verbose, strict=strict
    )
    if code:
        log.error(message)
        return 2

    return sys.exit(
        sig.weave(
            doc_root=doc,
            structure_name=structure,
            target_key=target,
            facet_key=facet,
            options=options,
            externals=EXTERNALS,
        )
    )


@app.command('changes')
def changes(  # noqa
    doc_root_pos: str = typer.Argument(''),
    doc_root: str = DocumentRoot,
    structure: str = StructureName,
    target: str = TargetName,
    facet: str = FacetName,
    verbose: bool = Verbosity,
    strict: bool = Strictness,
) -> int:
    """
    Weave in the changes for facet of target within document root.
    """
    code, message, doc, options = _verify_call_vector(
        doc_root=doc_root, doc_root_pos=doc_root_pos, verbose=verbose, strict=strict
    )
    if code:
        log.error(message)
        return 2

    return sys.exit(
        chg.weave(
            doc_root=doc,
            structure_name=structure,
            target_key=target,
            facet_key=facet,
            options=options,
            externals=EXTERNALS,
        )
    )


@app.command('concat')
def concat(  # noqa
    *,
    doc_root_pos: str = typer.Argument(''),
    doc_root: str = DocumentRoot,
    structure: str = StructureName,
    target: str = TargetName,
    facet: str = FacetName,
    verbose: bool = Verbosity,
    strict: bool = Strictness,
) -> int:
    """
    Concatenate the markdown tree for facet of target within render/pdf below document root.
    """
    code, message, doc, options = _verify_call_vector(
        doc_root=doc_root, doc_root_pos=doc_root_pos, verbose=verbose, strict=strict
    )
    if code:
        log.error(message)
        return 2

    return sys.exit(
        cat.concatenate(doc_root=doc, structure_name=structure, target_key=target, facet_key=facet, options=options)
    )


@app.command('render')
def render(  # noqa
    doc_root_pos: str = typer.Argument(''),
    doc_root: str = DocumentRoot,
    structure: str = StructureName,
    target: str = TargetName,
    facet: str = FacetName,
    label: str = LabelCall,
    verbose: bool = Verbosity,
    strict: bool = Strictness,
    patch_tables: bool = PatchTables,
    from_format_spec: str = FromFormatSpec,
    filter_cs_list: str = FilterCSList,
    approvals_strategy: str = ApprovalsStrategy,
) -> int:
    """
        Render the markdown tree for facet of target within render/pdf below document root.
    \n
        For ejected / customized templates set matching environment variables to the paths:\n
        - LIITOS_BOOKMATTER_TEMPLATE (for title page incl. approvals table)\n
        - LIITOS_PUBLISHER_TEMPLATE (for publisher page incl. changes and proprietary info)\n
        - LIITOS_METADATA_TEMPLATE (values to required known keys used on LaTeX level)\n
        - LIITOS_SETUP_TEMPLATE (general layout template)\n
        - DRIVER_TEMPLATE (template for general structure)
    """
    code, message, doc, options = _verify_call_vector(
        doc_root=doc_root,
        doc_root_pos=doc_root_pos,
        verbose=verbose,
        strict=strict,
        label=label,
        patch_tables=patch_tables,
        from_format_spec=from_format_spec,
        filter_cs_list=filter_cs_list,
        approvals_strategy=approvals_strategy,
    )
    if code:
        log.error(message)
        return sys.exit(code)

    start_time = dti.datetime.now(tz=dti.timezone.utc)
    start_ts = start_time.strftime(TS_FORMAT_PAYLOADS)
    log.info(f'Start timestamp ({start_ts})')
    code = cat.concatenate(doc_root=doc, structure_name=structure, target_key=target, facet_key=facet, options=options)
    if code:
        return sys.exit(code)

    idem = os.getcwd()
    doc = '../../'
    log.info(f'before met.weave(): {os.getcwd()} set doc ({doc})')
    externals = copy.deepcopy(EXTERNALS)
    code = met.weave(
        doc_root=doc, structure_name=structure, target_key=target, facet_key=facet, options=options, externals=externals
    )
    if code:
        return sys.exit(code)

    log.info(f'before sig.weave(): {os.getcwd()} set doc ({doc})')
    os.chdir(idem)
    log.info(f'relocated for sig.weave(): {os.getcwd()} with doc ({doc})')
    code = sig.weave(
        doc_root=doc, structure_name=structure, target_key=target, facet_key=facet, options=options, externals=externals
    )
    if code:
        return sys.exit(code)

    log.info(f'before chg.weave(): {os.getcwd()} set doc ({doc})')
    os.chdir(idem)
    log.info(f'relocated for chg.weave(): {os.getcwd()} with doc ({doc})')
    code = chg.weave(
        doc_root=doc, structure_name=structure, target_key=target, facet_key=facet, options=options, externals=externals
    )
    if code:
        return sys.exit(code)

    log.info(f'before chg.weave(): {os.getcwd()} set doc ({doc})')
    os.chdir(idem)
    log.info(f'relocated for chg.weave(): {os.getcwd()} with doc ({doc})')
    code = ren.der(doc_root=doc, structure_name=structure, target_key=target, facet_key=facet, options=options)

    end_time = dti.datetime.now(tz=dti.timezone.utc)
    end_ts = end_time.strftime(TS_FORMAT_PAYLOADS)
    duration_secs = (end_time - start_time).total_seconds()
    log.info(f'End timestamp ({end_ts})')
    acted = 'Rendered'
    if code == 0xFADECAFE:
        acted = 'Did not render'
        code = 0  # HACK A DID ACK
    log.info(f'{acted} {target} document for {facet} at {doc} in {duration_secs} secs')
    return sys.exit(code)


@app.command('report')
def report() -> int:
    """
    Report on the environment.
    """
    log.info(LOG_SEPARATOR)
    log.info('inspecting environment (tool version information):')
    for tool_key in TOOL_VERSION_COMMAND_MAP:
        too.report(tool_key)
    log.info(LOG_SEPARATOR)

    return sys.exit(0)


@app.command('eject')
def eject(  # noqa
    that: str = typer.Argument(''),
    out: str = OutputPath,
) -> int:
    """
    Eject a template. Enter unique part to retrieve, any unknown word to obtain the list of known templates.
    """
    return sys.exit(eje.this(thing=that, out=out))


@app.command('version')
def app_version() -> None:
    """
    Display the application version and exit.
    """
    callback(True)



"""Render the concat document to pdf."""

import json
import os
import pathlib
import re
import shutil
import time
from typing import Union, no_type_check

import yaml

import liitos.captions as cap
import liitos.concat as con
import liitos.description_lists as dsc
import liitos.figures as fig
import liitos.gather as gat
import liitos.labels as lab
import liitos.patch as pat
import liitos.tables as tab
import liitos.tools as too
from liitos import (
    CONTEXT,
    ENCODING,
    FROM_FORMAT_SPEC,
    LATEX_PAYLOAD_NAME,
    LOG_SEPARATOR,
    log,
    parse_csl,
)

DOC_BASE = pathlib.Path('..', '..')
STRUCTURE_PATH = DOC_BASE / 'structure.yml'
IMAGES_FOLDER = 'images/'
DIAGRAMS_FOLDER = 'diagrams/'
PATCH_SPEC_NAME = 'patch.yml'
INTER_PROCESS_SYNC_SECS = 0.1
INTER_PROCESS_SYNC_ATTEMPTS = 10
VENDORED_SVG_PAT = re.compile(r'^.+\]\([^.]+\.[^.]+\.svg\ .+$')


@no_type_check
def read_patches(folder_path: pathlib.Path, patches_path: pathlib.Path) -> tuple[list[tuple[str, str]], bool]:
    """Ja ja."""
    patches = []
    need_patching = False
    log.info(f'inspecting any patch spec file ({patches_path}) ...')
    if patches_path.is_file() and patches_path.stat().st_size:
        target_path = folder_path / PATCH_SPEC_NAME
        shutil.copy(patches_path, target_path)
        try:
            with open(patches_path, 'rt', encoding=ENCODING) as handle:
                patch_spec = yaml.safe_load(handle)
            need_patching = True
        except (OSError, UnicodeDecodeError) as err:
            log.error(f'failed to load patch spec from ({patches_path}) with ({err}) - patching will be skipped')
            need_patching = False
        if need_patching:
            try:
                patches = [(rep, lace) for rep, lace in patch_spec]
                patch_pair_count = len(patches)
                if not patch_pair_count:
                    need_patching = False
                    log.warning('- ignoring empty patch spec')
                else:
                    log.info(
                        f'- loaded {patch_pair_count} patch pair{"" if patch_pair_count == 1 else "s"}'
                        f' from patch spec file ({patches_path})'
                    )
            except ValueError as err:
                log.error(f'- failed to parse patch spec from ({patch_spec}) with ({err}) - patching will be skipped')
                need_patching = False
    else:
        if patches_path.is_file():
            log.warning(f'- ignoring empty patch spec file ({patches_path})')
        else:
            log.info(f'- no patch spec file ({patches_path}) detected')
    return patches, need_patching


@no_type_check
def der(
    doc_root: Union[str, pathlib.Path],
    structure_name: str,
    target_key: str,
    facet_key: str,
    options: dict[str, Union[bool, str]],
) -> int:
    """Later alligator."""
    log.info(LOG_SEPARATOR)
    log.info('entered render function ...')
    target_code = target_key
    facet_code = facet_key
    if not facet_code.strip() or not target_code.strip():
        log.error(f'render requires non-empty target ({target_code}) and facet ({facet_code}) codes')
        return 2
    log.info(f'parsed target ({target_code}) and facet ({facet_code}) from request')

    from_format_spec = options.get('from_format_spec', FROM_FORMAT_SPEC)
    filter_cs_list = parse_csl(options.get('filter_cs_list', ''))
    if filter_cs_list:
        log.info(f'parsed from-format-spec ({from_format_spec}) and filters ({", ".join(filter_cs_list)}) from request')
    else:
        log.info(f'parsed from-format-spec ({from_format_spec}) and no filters from request')

    structure, asset_map = gat.prelude(
        doc_root=doc_root, structure_name=structure_name, target_key=target_key, facet_key=facet_key, command='render'
    )
    log.info(f'prelude teleported processor into the document root at ({os.getcwd()}/)')

    rel_concat_folder_path = pathlib.Path('render/pdf/')
    rel_concat_folder_path.mkdir(parents=True, exist_ok=True)

    patches, need_patching = read_patches(rel_concat_folder_path, pathlib.Path(PATCH_SPEC_NAME))

    os.chdir(rel_concat_folder_path)
    log.info(f'render (this processor) teleported into the render/pdf location ({os.getcwd()}/)')

    log.info(LOG_SEPARATOR)
    log.info('Assessing the local version control status (compared to upstream) ...')
    too.ensure_separate_log_lines(too.vcs_probe)
    CONTEXT['builder_node_id'] = too.node_id()
    log.info('Context noted with:')
    log.info(f'- builder-node-id({CONTEXT.get("builder_node_id")})')
    log.info(f'- source-hash({CONTEXT.get("source_hash")})')
    log.info(f'- source-hint({CONTEXT.get("source_hint")})')

    ok, aspect_map = too.load_target(target_code, facet_code)
    if not ok or not aspect_map:
        return 0 if ok else 1

    do_render = aspect_map.get('render', None)
    if do_render is not None:
        log.info(f'found render instruction with value ({aspect_map["render"]})')

    if do_render is None or do_render:
        log.info('we will render ...')
    else:
        log.warning('we will not render ...')
        return 0xFADECAFE

    log.info(LOG_SEPARATOR)
    log.info('transforming SVG assets to high resolution PNG bitmaps ...')
    for path_to_dir in (IMAGES_FOLDER, DIAGRAMS_FOLDER):
        the_folder = pathlib.Path(path_to_dir)
        if not the_folder.is_dir():
            log.error(
                f'svg-to-png directory ({the_folder}) in ({pathlib.Path().cwd()}) does not exist or is no directory'
            )
            continue
        for svg in pathlib.Path(path_to_dir).iterdir():
            if svg.is_file() and svg.suffix == '.svg':
                png = str(svg).replace('.svg', '.png')
                svg_to_png_command = ['svgexport', svg, png, '100%']
                too.delegate(svg_to_png_command, 'svg-to-png')

    special_patching = []
    log.info(LOG_SEPARATOR)
    log.info('rewriting src attribute values of SVG to PNG sources ...')
    with open('document.md', 'rt', encoding=ENCODING) as handle:
        lines = [line.rstrip() for line in handle.readlines()]
    for slot, line in enumerate(lines):
        if line.startswith('![') and '](' in line:
            if VENDORED_SVG_PAT.match(line):
                if '.svg' in line and line.count('.') >= 2:
                    caption, src, alt, rest = con.parse_markdown_image(line)
                    stem, app_indicator, format_suffix = src.rsplit('.', 2)
                    log.info(f'- removing application indicator ({app_indicator}) from src ...')
                    if format_suffix != 'svg':
                        log.warning(f'  + format_suffix (.{format_suffix}) unexpected in <<{line.rstrip()}>> ...')
                    fine = f'![{caption}]({stem}.png "{alt}"){rest}'
                    log.info(f'  transform[#{slot + 1}]: {line}')
                    log.info(f'       into[#{slot + 1}]: {fine}')
                    lines[slot] = fine
                    dia_path_old = src.replace('.svg', '.png')
                    dia_path_new = f'{stem}.png'
                    dia_fine_rstrip = dia_path_new.rstrip()
                    if dia_path_old and dia_path_new:
                        special_patching.append((dia_path_old, dia_path_new))
                        log.info(
                            f'post-action[#{slot + 1}]: adding to queue for sync move: ({dia_path_old})'
                            f' -> ({dia_path_new})'
                        )
                    else:
                        log.warning(f'- old: {src.rstrip()}')
                        log.warning(f'- new: {dia_fine_rstrip}')
                    continue
            if '.svg' in line:
                fine = line.replace('.svg', '.png')
                log.info(f'  transform[#{slot + 1}]: {line}')
                log.info(f'       into[#{slot + 1}]: {fine}')
                lines[slot] = fine
                continue
    with open('document.md', 'wt', encoding=ENCODING) as handle:
        handle.write('\n'.join(lines))

    log.info(LOG_SEPARATOR)
    log.info('ensure diagram files can be found when patched ...')
    if special_patching:
        for old, mew in special_patching:
            source_asset = pathlib.Path(old)
            target_asset = pathlib.Path(mew)
            log.info(f'- moving: ({source_asset}) -> ({target_asset})')
            present = False
            remaining_attempts = INTER_PROCESS_SYNC_ATTEMPTS
            while remaining_attempts > 0 and not present:
                try:
                    present = source_asset.is_file()
                except Exception as ex:
                    log.error(f'    * probing for resource ({old}) failed with ({ex}) ... continuing')
                log.info(
                    f'  + resource ({old}) is{" " if present else " NOT "}present at ({source_asset})'
                    f' - attempt {11 - remaining_attempts} of {INTER_PROCESS_SYNC_ATTEMPTS} ...'
                )
                if present:
                    break
                time.sleep(INTER_PROCESS_SYNC_SECS)
                remaining_attempts -= 1
            if not source_asset.is_file():
                log.warning(
                    f'- resource ({old}) still not present at ({source_asset})'
                    f' as seen from ({os.getcwd()}) after {remaining_attempts} attempts'
                    f' and ({round(remaining_attempts * INTER_PROCESS_SYNC_SECS, 0) :.0f} seconds waiting)'
                )
            shutil.move(source_asset, target_asset)
    else:
        log.info('post-action queue (from reference renaming) is empty - nothing to move')
    log.info(LOG_SEPARATOR)

    # prototyping >>>
    fmt_spec = from_format_spec
    in_doc = 'document.md'
    out_doc = 'ast-no-filter.json'
    markdown_to_ast_command = [
        'pandoc',
        '--verbose',
        '-f',
        fmt_spec,
        '-t',
        'json',
        in_doc,
        '-o',
        out_doc,
    ]
    log.info(LOG_SEPARATOR)
    log.info(f'executing ({" ".join(markdown_to_ast_command)}) ...')
    if code := too.delegate(markdown_to_ast_command, 'markdown-to-ast'):
        return code

    log.info(LOG_SEPARATOR)

    mermaid_caption_map = too.mermaid_captions_from_json_ast(out_doc)
    log.info(LOG_SEPARATOR)
    # no KISS too.ensure_separate_log_lines(json.dumps, [mermaid_caption_map, 2])
    for line in json.dumps(mermaid_caption_map, indent=2).split('\n'):
        for fine in line.split('\n'):
            log.info(fine)
    log.info(LOG_SEPARATOR)

    # <<< prototyping

    fmt_spec = from_format_spec
    in_doc = 'document.md'
    out_doc = LATEX_PAYLOAD_NAME
    markdown_to_latex_command = [
        'pandoc',
        '--verbose',
        '-f',
        fmt_spec,
        '-t',
        'latex',
        in_doc,
        '-o',
        out_doc,
    ]
    if filter_cs_list:
        filters = [added_prefix for expr in filter_cs_list for added_prefix in ('--filter', expr)]
        markdown_to_latex_command += filters
    log.info(LOG_SEPARATOR)
    log.info(f'executing ({" ".join(markdown_to_latex_command)}) ...')
    if code := too.delegate(markdown_to_latex_command, 'markdown-to-latex'):
        return code

    log.info(LOG_SEPARATOR)
    log.info(f'load text lines from intermediate {LATEX_PAYLOAD_NAME} file before internal transforms ...')
    with open(LATEX_PAYLOAD_NAME, 'rt', encoding=ENCODING) as handle:
        lines = [line.rstrip() for line in handle.readlines()]

    lines = too.execute_filter(
        cap.weave,
        head='move any captions below tables ...',
        backup='document-before-caps-patch.tex.txt',
        label='captions-below-tables',
        text_lines=lines,
        lookup=None,
    )

    lines = too.execute_filter(
        lab.inject,
        head='inject stem (derived from file name) labels ...',
        backup='document-before-inject-stem-label-patch.tex.txt',
        label='inject-stem-derived-labels',
        text_lines=lines,
        lookup=mermaid_caption_map,
    )

    lines = too.execute_filter(
        fig.scale,
        head='scale figures ...',
        backup='document-before-scale-figures-patch.tex.txt',
        label='inject-scale-figures',
        text_lines=lines,
        lookup=None,
    )

    lines = too.execute_filter(
        dsc.options,
        head='add options to descriptions (definition lists) ...',
        backup='document-before-description-options-patch.tex.txt',
        label='inject-description-options',
        text_lines=lines,
        lookup=None,
    )

    if options.get('patch_tables', False):
        lines = too.execute_filter(
            tab.patch,
            head='patching tables EXPERIMENTAL (table-shape) ...',
            backup='document-before-table-shape-patch.tex.txt',
            label='changed-table-shape',
            text_lines=lines,
            lookup=None,
        )
    else:
        log.info(LOG_SEPARATOR)
        log.info('not patching tables but commenting out (ignoring) any columns command (table-shape) ...')
        patched_lines = [f'%IGNORED_{v}' if v.startswith(r'\columns=') else v for v in lines]
        patched_lines = [f'%IGNORED_{v}' if v.startswith(r'\tablefontsize=') else v for v in patched_lines]
        log.info('diff of the (ignore-table-shape-if-not-patched) filter result:')
        too.log_unified_diff(lines, patched_lines)
        lines = patched_lines
        log.info(LOG_SEPARATOR)

    if need_patching:
        log.info(LOG_SEPARATOR)
        log.info('apply user patches ...')
        doc_before_user_patch = 'document-before-user-patch.tex.txt'
        with open(doc_before_user_patch, 'wt', encoding=ENCODING) as handle:
            handle.write('\n'.join(lines))
        patched_lines = pat.apply(patches, lines)
        with open(LATEX_PAYLOAD_NAME, 'wt', encoding=ENCODING) as handle:
            handle.write('\n'.join(patched_lines))
        log.info('diff of the (user-patches) filter result:')
        too.log_unified_diff(lines, patched_lines)
        lines = patched_lines
    else:
        log.info(LOG_SEPARATOR)
        log.info('skipping application of user patches ...')

    log.info(LOG_SEPARATOR)
    log.info(f'Internal text line buffer counts {len(lines)} lines')

    log.info(LOG_SEPARATOR)
    log.info('cp -a driver.tex this.tex ...')
    source_asset = 'driver.tex'
    target_asset = 'this.tex'
    shutil.copy(source_asset, target_asset)

    latex_to_pdf_command = ['lualatex', '--shell-escape', 'this.tex']
    log.info(LOG_SEPARATOR)
    log.info('1/3) lualatex --shell-escape this.tex ...')
    if code := too.delegate(latex_to_pdf_command, 'latex-to-pdf(1/3)'):
        return code

    log.info(LOG_SEPARATOR)
    log.info('2/3) lualatex --shell-escape this.tex ...')
    if code := too.delegate(latex_to_pdf_command, 'latex-to-pdf(2/3)'):
        return code

    log.info(LOG_SEPARATOR)
    log.info('3/3) lualatex --shell-escape this.tex ...')
    if code := too.delegate(latex_to_pdf_command, 'latex-to-pdf(3/3)'):
        return code

    if str(options.get('label', '')).strip():
        labeling_call = str(options['label']).strip().split()
        labeling_call.extend(
            [
                '--key-value-pairs',
                (
                    f'BuilderNodeID={CONTEXT["builder_node_id"]}'
                    f',SourceHash={CONTEXT.get("source_hash", "no-source-hash-given")}'
                    f',SourceHint={CONTEXT.get("source_hint", "no-source-hint-given")}'
                ),
            ]
        )
        log.info(LOG_SEPARATOR)
        log.info(f'Labeling the resulting pdf file per ({" ".join(labeling_call)})')
        too.delegate(labeling_call, 'label-pdf')
        log.info(LOG_SEPARATOR)

    log.info(LOG_SEPARATOR)
    log.info('Moving stuff around (result phase) ...')
    source_asset = 'this.pdf'
    target_asset = '../index.pdf'
    shutil.copy(source_asset, target_asset)

    log.info(LOG_SEPARATOR)
    log.info('Deliverable taxonomy: ...')
    too.report_taxonomy(pathlib.Path(target_asset))

    pdffonts_command = ['pdffonts', target_asset]
    too.delegate(pdffonts_command, 'assess-pdf-fonts')

    log.info(LOG_SEPARATOR)
    log.info('done.')
    log.info(LOG_SEPARATOR)

    return 0



"""Apply all pairs in patch module on document."""

import re
from collections.abc import Iterable, Iterator
from typing import Union, no_type_check

from liitos import log

# The "target pattern for a line base minimal regex parser"
_ = r"""
\columns=

\begin{longtable}[]{
... \real{fwidth_1}}
...
... \real{fwidth_n}}@{}}
\toprule\noalign{}
\begin{minipage}[
text_1
\end{minipage} & \begin{minipage}[
text_2
...
\end{minipage} & \begin{minipage}[
text_n
\end{minipage} \\
\midrule\noalign{}
\endfirsthead
\toprule\noalign{}
\begin{minipage}[
text_1
\end{minipage} & \begin{minipage}[
text_2
...
\end{minipage} & \begin{minipage}[
text_n
\end{minipage} \\
\midrule\noalign{}
\endhead
\bottomrule\noalign{}
\endlastfoot
cell_1_1 & cell_1_2 & ... & cell_1_n \\
cell_1_2 & cell_2_2 & ... & cell_2_n \\
...
row_1_m & row_2_m & ... & cell_n_m \\
\rowcolor{white}
\caption{cap_text_x
cap_text_y\tabularnewline
\end{longtable}
"""

TAB_START_TOK = r'\begin{longtable}[]{'  # '@{}'
TOP_RULE = r'\toprule()'
MID_RULE = r'\midrule()'
END_HEAD = r'\endhead'
END_DATA_ROW = r'\\'
BOT_RULE = r'\bottomrule()'
TAB_END_TOK = r'\end{longtable}'

TAB_NEW_START = r"""\begin{small}
\begin{longtable}[]{|
>{\raggedright\arraybackslash}p{(\columnwidth - 12\tabcolsep) * \real{0.1500}}|
>{\raggedright\arraybackslash}p{(\columnwidth - 12\tabcolsep) * \real{0.5500}}|
>{\raggedright\arraybackslash}p{(\columnwidth - 12\tabcolsep) * \real{0.1500}}|
>{\raggedright\arraybackslash}p{(\columnwidth - 12\tabcolsep) * \real{0.2000}}|}
\hline"""

TAB_HACKED_HEAD = r"""\begin{minipage}[b]{\linewidth}\raggedright
\ \mbox{\textbf{Key}}
\end{minipage} & \begin{minipage}[b]{\linewidth}\raggedright
\mbox{\textbf{Summary}}
\end{minipage} & \begin{minipage}[b]{\linewidth}\raggedright
\mbox{\textbf{Parent}} \mbox{\textbf{Requirement}}
\end{minipage} & \begin{minipage}[b]{\linewidth}\raggedright
\mbox{\textbf{Means of}} \mbox{\textbf{Compliance (MOC)}}
\end{minipage} \\
\hline
\endfirsthead
\multicolumn{4}{@{}l}{\small \ldots continued}\\\hline
\hline
\begin{minipage}[b]{\linewidth}\raggedright
\ \mbox{\textbf{Key}}
\end{minipage} & \begin{minipage}[b]{\linewidth}\raggedright
\mbox{\textbf{Summary}}
\end{minipage} & \begin{minipage}[b]{\linewidth}\raggedright
\mbox{\textbf{Parent}} \mbox{\textbf{Requirement}}
\end{minipage} & \begin{minipage}[b]{\linewidth}\raggedright
\mbox{\textbf{Means of}} \mbox{\textbf{Compliance (MOC)}}
\end{minipage} \\
\endhead
\hline"""

NEW_RULE = r'\hline'

TAB_NEW_END = r"""\end{longtable}
\end{small}
\vspace*{-2em}
\begin{footnotesize}
ANNOTATION
\end{footnotesize}"""

COMMA = ','


class Table:
    """Some adhoc structure to encapsulate the source and target table."""

    SourceMapType = list[tuple[int, str]]
    ColumnsType = dict[str, dict[str, Union[float, int, str]]]

    # ---- begin of LBP skeleton / shape ---
    LBP_STARTSWITH_TAB_ENV_BEGIN = r'\begin{longtable}[]{'
    LBP_REAL_INNER_COLW_PAT = re.compile(r'^(?P<clspec>.+)\\real{(?P<cwval>[0-9.]+)}}\s*$')
    LBP_REAL_OUTER_COLW_PAT = re.compile(r'^(?P<clspec>.+)\\real{(?P<cwval>[0-9.]+)}}@{}}\s*$')
    # Width lines for header:
    FUT_LSPLIT_ONCE_FOR_PREFIX_VAL_COMMA_RIGHT = '}}'
    FUT_LSPLIT_ONCE_FOR_PREFIX_COMMA_VAL = r'\real{'
    # then concat PREFIX + r'\real{' + str(column_width_new) + '}}' + RIGHT

    LBP_TOP_RULE_CONTEXT_STARTSWITH = r'\toprule\noalign{}'
    LPB_START_COLUMN_LABEL_STARTSWITH = r'\begin{minipage}['
    LBP_SEP_COLUMN_LABEL_STARTSWITH = r'\end{minipage} & \begin{minipage}['
    LBP_STOP_COLUMN_LABEL_STARTSWITH = r'\end{minipage} \\'

    LBP_MID_RULE_CONTEXT_STARTSWITH = r'\midrule\noalign{}'

    LBP_END_FIRST_HEAD_STARTSWITH = r'\endfirsthead'

    # LBP_TOP_RULE_CONTEXT_STARTSWITH = r'\toprule\noalign{}'
    # LPB_START_COLUMN_LABEL_STARTSWITH = r'\begin{minipage}['
    # LBP_SEP_COLUMN_LABEL_STARTSWITH = r'\end{minipage} & \begin{minipage}['
    # LBP_STOP_COLUMN_LABEL_STARTSWITH = r'\end{minipage} \\'

    # LBP_MID_RULE_CONTEXT_STARTSWITH = r'\midrule\noalign{}'

    LBP_END_ALL_HEAD_STARTSWITH = r'\endhead'

    LBP_BOTTOM_RULE_CONTEXT_STARTSWITH = r'\bottomrule\noalign{}'

    LBP_END_LAST_FOOT_STARTSWITH = r'\endlastfoot'

    # ... data lines - we want inject of r'\hline' following every data line (not text line)
    # -> that is, inject after lines ending with r'\\'

    LBP_END_OF_DATA_STARTSWITH = r'\rowcolor{white}'
    LBP_START_CAP_STARTSWITH = r'\caption{'
    LBP_STOP_CAP_ENDSWITH = r'\tabularnewline'
    LBP_STARTSWITH_TAB_ENV_END = r'\end{longtable}'

    # ---- end of LBP skeleton / shape ---
    @no_type_check
    def __init__(self, anchor: int, start_line: str, text_lines: Iterator[str], widths: list[float], font_sz: str = ''):
        """Initialize the table from source text lines anchored at anchor.
        The implementation allows reuse of the iterator on caller site for extracting subsequent tables in one go.
        """
        self.src_map: Table.SourceMapType = [(anchor, start_line.rstrip())]
        self.data_row_ends: Table.SourceMapType = []
        self.columns: Table.ColumnsType = {}
        self.target_widths: list[float] = widths
        self.source_widths: list[float] = []
        self.font_size = font_sz
        log.info(f'Received {anchor=}, {start_line=}, target {widths=}, and {font_sz=}')
        local_number = 0
        consumed = False
        while not consumed:
            local_number += 1
            pos = local_number + anchor
            line = next(text_lines).rstrip()
            self.src_map.append((pos, line))
            if line.startswith(Table.LBP_STARTSWITH_TAB_ENV_END):
                consumed = True

        self.parse_columns()
        self.parse_data_rows()
        self.data_row_count = len(self.data_row_ends)
        self.cw_patches: dict[str, str] = {}
        self.create_width_patches()
        log.info(f'Parsed {len(self.target_widths)} x {self.data_row_count} table starting at anchor {anchor}')

    @no_type_check
    def create_width_patches(self):
        """If widths are meaningful and consistent create the patches with the zero-based line-numbers as keys."""
        if not self.source_widths:
            log.warning('Found no useful width information')
            return {}
        wrapper = r'\real{'
        postfix = '}}'
        finalize = '@{}}'
        ranks = list(self.columns)
        for rank in ranks:
            anchor_str = str(self.columns[rank]['col_spec_line'])
            prefix = self.columns[rank]['colspec_prefix']
            value = self.columns[rank]['width']
            # concat PREFIX + r'\real{' + str(column_width_new) + '}}'
            self.cw_patches[anchor_str] = prefix + wrapper + str(value) + postfix
            if rank == ranks[-1]:
                self.cw_patches[anchor_str] += finalize

    def width_patches(self) -> dict[str, str]:
        """Return the map of width patches with the zero-based line-numbers as keys."""
        return self.cw_patches

    def source_map(self) -> SourceMapType:
        """Return the source map data (a random accessible sequence of pairs) mapping abs line number to text line."""
        return self.src_map

    def column_data(self) -> ColumnsType:
        """Return the column data (an ordered dict of first labels, other labels, and widths) with abs line map."""
        return self.columns

    def column_source_widths(self) -> list[float]:
        """Return the incoming column widths."""
        return self.source_widths

    def column_target_widths(self) -> list[float]:
        """Return the outgoing column widths."""
        return self.target_widths

    @no_type_check
    def table_width(self) -> float:
        """Return the sum of all column widths."""
        return sum(self.columns[r].get('width', 0) for r in self.columns)

    def data_row_seps(self) -> SourceMapType:
        """Return the map to the data row ends for injecting separators."""
        return self.data_row_ends

    @no_type_check
    def transform_widths(self) -> None:
        """Apply the target transform to column widths."""
        self.source_widths = [self.columns[rank]['width'] for rank in self.columns]
        if not self.target_widths:
            log.info('No target widths given - maintaining source column widths')
            self.target_widths = self.source_widths
            return
        if len(self.target_widths) != len(self.source_widths):
            log.warning(
                f'Mismatching {len(self.target_widths)} target widths given - maintaining'
                f'the {len(self.source_widths)} source column widths'
            )
            self.target_widths = self.source_widths
            return
        log.info('Applying target widths given - adapting source column widths')
        for rank, target_width in zip(self.columns, self.target_widths):
            self.columns[rank]['width'] = target_width

    @no_type_check
    def parse_columns(self) -> None:
        """Parse the head to extract the columns."""
        self.parse_column_widths()
        self.parse_column_first_head()
        self.parse_column_other_head()
        self.transform_widths()

    def parse_column_widths(self) -> None:
        r"""Parse the column width declarations to initialize the columns data.

        \begin{longtable}[]{@{}%wun-based-line-9
          >{\raggedright\arraybackslash}p{(\columnwidth - 6\tabcolsep) * \real{0.1118}}
          >{\raggedright\arraybackslash}p{(\columnwidth - 6\tabcolsep) * \real{0.5776}}
          >{\raggedright\arraybackslash}p{(\columnwidth - 6\tabcolsep) * \real{0.1739}}
          >{\raggedright\arraybackslash}p{(\columnwidth - 6\tabcolsep) * \real{0.1366}}@{}}
        \toprule\noalign{}
        """
        rank = 0
        for anchor, text in self.src_map:
            if text.startswith(Table.LBP_STARTSWITH_TAB_ENV_BEGIN):
                continue
            if text.startswith(Table.LBP_TOP_RULE_CONTEXT_STARTSWITH):
                break
            m = Table.LBP_REAL_INNER_COLW_PAT.match(text)
            if m:
                self.columns[str(rank)] = {
                    'first_label': '',
                    'first_label_line': -1,
                    'continued_label': '',
                    'continued_label_line': -1,
                    'col_spec_line': anchor,
                    'colspec_prefix': m.groupdict()['clspec'],
                    'width': float(m.groupdict()['cwval']),
                }
                rank += 1
                continue
            m = Table.LBP_REAL_OUTER_COLW_PAT.match(text)
            if m:
                self.columns[str(rank)] = {
                    'first_label': '',
                    'first_label_line': -1,
                    'continued_label': '',
                    'continued_label_line': -1,
                    'col_spec_line': anchor,
                    'colspec_prefix': m.groupdict()['clspec'],
                    'width': float(m.groupdict()['cwval']),
                }
                rank += 1
                continue

    def parse_column_first_head(self) -> None:
        r"""Parse the head to extract the columns.

        \begin{minipage}[b]{\linewidth}\raggedright
        Parameter
        \end{minipage} & \begin{minipage}[b]{\linewidth}\raggedright
        Description
        \end{minipage} & \begin{minipage}[b]{\linewidth}\raggedright
        Name
        \end{minipage} & \begin{minipage}[b]{\linewidth}\raggedright
        Example
        \end{minipage} \\
        \midrule\noalign{}
        \endfirsthead
        """
        rank = 0
        first_head = False
        label_next = False
        for anchor, text in self.src_map:
            if text.startswith(Table.LBP_TOP_RULE_CONTEXT_STARTSWITH):
                first_head = True
                continue
            if not first_head:
                continue
            if text.startswith(Table.LBP_END_FIRST_HEAD_STARTSWITH):
                break
            if text.startswith(Table.LPB_START_COLUMN_LABEL_STARTSWITH):
                label_next = True
                continue
            if text.startswith(Table.LBP_SEP_COLUMN_LABEL_STARTSWITH):
                label_next = True
                continue
            if text.startswith(Table.LBP_STOP_COLUMN_LABEL_STARTSWITH):
                label_next = True
                continue
            if label_next:
                self.columns[str(rank)]['first_label'] = text.strip()
                self.columns[str(rank)]['first_label_line'] = anchor
                rank += 1
                if str(rank) in self.columns:
                    continue
                break

    def parse_column_other_head(self) -> None:
        r"""Parse the other heads to extract the column labelss.

        \endfirsthead
        \toprule\noalign{}
        \begin{minipage}[b]{\linewidth}\raggedright
        Parameter
        \end{minipage} & \begin{minipage}[b]{\linewidth}\raggedright
        Description
        \end{minipage} & \begin{minipage}[b]{\linewidth}\raggedright
        Name
        \end{minipage} & \begin{minipage}[b]{\linewidth}\raggedright
        Example
        \end{minipage} \\
        \midrule\noalign{}
        \endhead
        """
        rank = 0
        continued_head = False
        label_next = False
        for anchor, text in self.src_map:
            if text.startswith(Table.LBP_END_FIRST_HEAD_STARTSWITH):
                continued_head = True
                continue
            if not continued_head:
                continue
            if text.startswith(Table.LBP_END_ALL_HEAD_STARTSWITH):
                break
            if text.startswith(Table.LPB_START_COLUMN_LABEL_STARTSWITH):
                label_next = True
                continue
            if text.startswith(Table.LBP_SEP_COLUMN_LABEL_STARTSWITH):
                label_next = True
                continue
            if text.startswith(Table.LBP_STOP_COLUMN_LABEL_STARTSWITH):
                label_next = True
                continue
            if label_next:
                self.columns[str(rank)]['continued_label'] = text.strip()
                self.columns[str(rank)]['continued_label_line'] = anchor
                rank += 1
                if str(rank) in self.columns:
                    continue
                break

    def parse_data_rows(self) -> None:
        r"""Parse the data rows.

        \endlastfoot
        A2 & B2 & C2 & D2 \\
        \end{longtable}
        """
        data_section = False
        for anchor, text in self.src_map:
            if text.startswith(Table.LBP_END_LAST_FOOT_STARTSWITH):
                data_section = True
                continue
            if text.startswith(Table.LBP_STARTSWITH_TAB_ENV_END):
                break
            if data_section and r'\\' in text:
                self.data_row_ends.append((anchor, text))
                continue


def parse_table_font_size_command(slot: int, text_line: str) -> tuple[bool, str, str]:
    """Parse the \\tablefontsize=footnotesize command."""
    backslash = '\\'
    known_sizes = (
        'tiny',
        'scriptsize',
        'footnotesize',
        'small',
        'normalsize',
        'large',
        'Large',
        'LARGE',
        'huge',
        'Huge',
    )
    if text_line.startswith(r'\tablefontsize='):
        log.info(f'trigger a fontsize mod for the next table environment at line #{slot + 1}|{text_line}')
        try:
            font_size = text_line.split('=', 1)[1].strip()  # r'\tablefontsize=Huge'  --> 'Huge'
            if font_size.startswith(backslash):
                font_size = font_size.lstrip(backslash)
            if font_size not in known_sizes:
                log.error(f'failed to map given fontsize ({font_size}) into known sizes ({",".join(known_sizes)})')
                return False, text_line, ''
            log.info(f' -> parsed table fontsize mod as ({font_size})')
            return True, '', font_size
        except Exception as err:
            log.error(f'failed to parse table fontsize value from {text_line.strip()} with err: {err}')
            return False, text_line, ''
    else:
        return False, text_line, ''


def parse_columns_command(slot: int, text_line: str) -> tuple[bool, str, list[float]]:
    """Parse the \\columns=,0.2,0.7 command."""
    if text_line.startswith(r'\columns='):
        log.info(f'trigger a columns mod for the next table environment at line #{slot + 1}|{text_line}')
        try:
            cols_csv = text_line.split('=', 1)[1].strip()  # r'\columns    =    , 20\%,70\%'  --> r', 20\%,70\%'
            cols = [v.strip() for v in cols_csv.split(COMMA)]
            widths = [float(v.replace(r'\%', '')) / 100 if r'\%' in v else (float(v) if v else 0) for v in cols]
            rest = round(1 - sum(round(w, 5) for w in widths), 5)
            widths = [v if v else rest for v in widths]
            log.info(f' -> parsed columns mod as | {" | ".join(str(round(v, 2)) for v in widths)} |')
            return True, '', widths
        except Exception as err:
            log.error(f'failed to parse columns values from {text_line.strip()} with err: {err}')
            return False, text_line, []
    else:
        return False, text_line, []


@no_type_check
def patch(incoming: Iterable[str], lookup: Union[dict[str, str], None] = None) -> list[str]:
    """Later alligator. \\columns=,0.2,0.7 as mandatory trigger"""
    table_section, head, annotation = False, False, False
    table_ranges = []
    guess_slot = 0
    table_range = {}
    has_column = False
    has_font_size = False
    widths: list[float] = []
    font_size = ''
    comment_outs = []
    for n, text in enumerate(incoming):
        if not table_section:
            if not has_font_size:
                has_font_size, text_line, font_size = parse_table_font_size_command(n, text)
                if has_font_size:
                    comment_outs.append(n)
            if not has_font_size:
                continue

            if not has_column:
                has_column, text_line, widths = parse_columns_command(n, text)
                if has_column:
                    comment_outs.append(n)
            if not has_column:
                continue

            if not text.startswith(TAB_START_TOK):
                continue
            table_range['start'] = n
            table_section = True
            head = True
            table_range['end_data_row'] = []
            continue

        if text.startswith(TOP_RULE):
            table_range['top_rule'] = n
            continue

        if text.startswith(MID_RULE):
            table_range['mid_rule'] = n
            continue

        if text.startswith(END_HEAD):
            table_range['end_head'] = n
            head = False
            continue

        if not head and text.strip().endswith(END_DATA_ROW):
            table_range['end_data_row'].append(n)
            continue

        if text.startswith(BOT_RULE):
            table_range['bottom_rule'] = n
            continue

        if text.startswith(TAB_END_TOK):
            table_range['end'] = n
            annotation = True
            guess_slot = n + 2
            continue

        if annotation and n == guess_slot:
            table_range['amend'] = n
            table_ranges.append(table_range)
            table_range = {}
            annotation, table_section = False, False

    log.info(f'Detected {len(table_ranges)} tables (method from before version 2023.2.12):')
    for thing in table_ranges:
        log.info(f'- {thing}')

    tables_in, on_off_slots = [], []
    for table in table_ranges:
        from_here = table['start']
        thru_there = table['amend']
        log.info('Table:')
        log.info(f'-from {incoming[from_here]}')
        log.info(f'-thru {incoming[thru_there]}')
        on_off = (from_here, thru_there + 1)
        on_off_slots.append(on_off)
        tables_in.append((on_off, [line for line in incoming[on_off[0] : on_off[1]]]))

    log.debug('# - - - 8< - - -')
    if tables_in:
        log.debug(str('\n'.join(tables_in[0][1])))
    log.debug('# - - - 8< - - -')

    reader = iter(incoming)
    tables = []
    comment_outs = []
    n = 0
    widths = []
    font_size = ''
    for line in reader:
        log.debug(f'zero-based-line-no={n}, text=({line}) table-count={len(tables)}')
        if not line.startswith(Table.LBP_STARTSWITH_TAB_ENV_BEGIN):
            if line.startswith(r'\tablefontsize='):
                has_font_size, text_line, font_size = parse_table_font_size_command(n, line)
                log.info(f'    + {has_font_size=}, {text_line=}, {font_size=}')
                if has_font_size:
                    comment_outs.append(n)
                    log.info(f'FONT-SIZE at <<{n}>>')
            if line.startswith(r'\columns='):
                has_column, text_line, widths = parse_columns_command(n, line)
                log.info(f'    + {has_column=}, {text_line=}, {widths=}')
                if has_column:
                    comment_outs.append(n)
                    log.info(f'COLUMNS-WIDTH at <<{n}>>')
            n += 1
        else:
            table = Table(n, line, reader, widths, font_size)  # sharing the meal - instead of iter(lines_buffer[n:]))
            tables.append(table)
            n += len(tables[-1].source_map())
            log.debug(f'- incremented n to {n}')
            log.debug(f'! next n (zero offset) is {n}')
            widths = []
            font_size = ''

    log.info('---')
    for n, table in enumerate(tables, start=1):
        log.info(f'Table #{n} (total width = {table.table_width()}):')
        for rank, column in table.column_data().items():
            log.info(f'{rank} -> {column}')
        log.info(f'- source widths = {table.column_source_widths()}):')
        log.info(f'- target widths = {table.column_target_widths()}):')
        for numba, replacement in table.width_patches().items():
            log.info(f'{numba} -> {replacement}')
        for anchor, text in table.data_row_seps():
            log.info(f'{anchor} -> {text}')
        log.info(f'= (fontsize command = "{table.font_size}"):')
    log.info('---')
    log.info(f'Comment out the following {len(comment_outs)} lines (zero based numbers) - punch:')
    for number in comment_outs:
        log.info(f'- {number}')

    wideners = {}
    for table in tables:
        for numba, replacement in table.width_patches().items():
            wideners[numba] = replacement
    widen_me = set(wideners)
    log.debug('widen me has:')
    log.debug(list(widen_me))
    log.debug('--- global replacement width lines: ---')
    for numba, replacement in wideners.items():
        log.debug(f'{numba} => {replacement}')
    log.debug('---')

    sizers = {}
    for table in tables:
        if table.font_size:
            sizers[str(table.src_map[0][0])] = '\\begin{' + table.font_size + '}\n' + table.src_map[0][1]
            sizers[str(table.src_map[-1][0])] = table.src_map[-1][1] + '\n' + '\\end{' + table.font_size + '}'
    size_me = set(sizers)
    log.debug('size me has:')
    log.debug(list(size_me))
    log.debug('--- global replacement sizer lines: ---')
    for numba, replacement in sizers.items():
        log.debug(f'{numba} => {replacement}')
    log.debug('---')

    out = []
    # next_slot = 0
    punch_me = set(comment_outs)
    for n, line in enumerate(incoming):
        if n in punch_me:
            corrected = f'%CONSIDERED_{line}'
            out.append(corrected)
            log.info(f' (x) Punched out line {n} -> ({corrected})')
            continue
        if str(n) in widen_me:
            out.append(wideners[str(n)])
            log.info(f' (<) Incoming: ({line})')
            log.info(f' (>) Outgoing: ({wideners[str(n)]})')
            continue
        if str(n) in size_me:
            out.append(sizers[str(n)])
            log.info(f' (<) Incoming: ({line})')
            log.info(f' (>) Outgoing: ({sizers[str(n)]})')
            continue
        out.append(line)

        # if next_slot < len(on_off_slots):
        #     trigger_on, trigger_off = on_off_slots[next_slot]
        #     tb = table_ranges[next_slot]
        # else:
        #     trigger_on = None
        # if trigger_on is None:
        #     out.append(line)
        #     continue
        #
        # if n < trigger_on:
        #     out.append(line)
        #     continue
        # if n == trigger_on:
        #     out.append(TAB_NEW_START)
        #     out.append(TAB_HACKED_HEAD)
        #     continue
        # if n <= tb['end_head']:
        #     continue
        # if n < tb.get('bottom_rule', 0):
        #     out.append(line)
        #     if n in tb['end_data_row']:  # type: ignore
        #         out.append(NEW_RULE)
        #     continue
        # if tb.get('bottom_rule', 0) <= n < tb['amend']:
        #     continue
        # if n == tb['amend']:
        #     out.append(TAB_NEW_END.replace('ANNOTATION', line))
        #     next_slot += 1

    log.warning('Disabled naive table patching from before version 2023.2.12 for now')

    log.debug(' -----> ')
    log.debug('# - - - 8< - - -')
    log.debug(str('\n'.join(out)))
    log.debug('# - - - 8< - - -')

    return out



from collections.abc import Iterable
from typing import Union

from liitos import log

NO_RESCALE: Union[float, int] = 0


def scale(incoming: Iterable[str], lookup: Union[dict[str, str], None] = None) -> list[str]:
    """Later alligator."""
    outgoing = []
    modus = 'copy'
    rescale = NO_RESCALE
    for slot, line in enumerate(incoming):
        if modus == 'copy':
            if line.startswith(r'\scale='):
                log.info(f'trigger a scale mod for the next figure environment at line #{slot + 1}|{line}')
                modus = 'scale'
                scale = line  # only for reporting wil not pass the filter
                try:
                    sca = scale.split('=', 1)[1].strip()  # \scale    =    75\%  --> 75\%
                    rescale = float(sca.replace(r'\%', '')) / 100 if r'\%' in sca else float(sca)
                except Exception as err:
                    log.error(f'failed to parse scale value from {line.strip()} with err: {err}')
            else:
                outgoing.append(line)

        else:  # if modus == 'scale':
            if line.startswith(r'\includegraphics{'):
                if rescale != NO_RESCALE:
                    log.info(f'- found the scale target start at line #{slot + 1}|{line}')
                    target = line.replace(r'\includegraphics', '')
                    option = (
                        f'[width={round(rescale, 2)}\\textwidth,height={round(rescale, 2)}'
                        '\\textheight,keepaspectratio]'
                    )
                    outgoing.append(f'\\includegraphics{option}{target}')
                else:
                    outgoing.append(line)
                modus = 'copy'
                rescale = NO_RESCALE
            elif r'\pandocbounded{\includegraphics' in line:
                if rescale != NO_RESCALE:
                    log.info(f'- found the scale target start at line #{slot + 1}|{line}')
                    target = line.replace(r'\pandocbounded{\includegraphics', '').replace('[keepaspectratio]', '')
                    parts = target.split('}}')
                    rest = ''
                    if len(parts) >= 1:
                        inside = parts[0] + '}'
                        if len(parts) == 2:
                            rest = parts[1].lstrip('}')
                    option = (
                        f'[width={round(rescale, 2)}\\textwidth,height={round(rescale, 2)}'
                        '\\textheight,keepaspectratio]'
                    )
                    outgoing.append(f'\\pandocbounded{{\\includegraphics{option}{inside}}}{rest}')
                else:
                    outgoing.append(line)
                modus = 'copy'
                rescale = NO_RESCALE
            else:
                outgoing.append(line)

    return outgoing



"""Weave the content of the changes data file into the output structure (for now LaTeX).

# Supported Table Layouts

# Layout `named` (the old default)

| Iss. | Rev. | Author           | Description                                 |
|:-----|:-----|:-----------------|:--------------------------------------------|
| 001  | 00   | Ann Author       | Initial issue                               |
| 001  | 01   | Another Author   | The mistakes were fixed                     |
| 002  | 00   | That was them    | Other mistakes, maybe, who knows, not ours? |

Table: The named table is simple to grow by appending rows.

The named layout relies on the skeleton in the pubisher.tex.in template.

# Layout `anonymous` (the new default)

| Iss. | Rev. | Description                                                    |
|:-----|:-----|:---------------------------------------------------------------|
| 001  | 00   | Initial issue                                                  |
| 001  | 01   | The mistakes were fixed                                        |
| 002  | 00   | Other mistakes, maybe, who knows, not ours?                    |

Table: This anonymous table is also simple to grow by appending rows.

The anonymous layout requires more dynamic LaTeX generation and thus generates the construct
from the data inside this module.
"""

import pathlib
from typing import Generator, Union, no_type_check

import liitos.gather as gat
import liitos.template as tpl
import liitos.tools as too
from liitos import ENCODING, ExternalsType, LOG_SEPARATOR, PathLike, log

CHANGE_ROW_TOKEN = r'THE.ISSUE.CODE & THE.REVISION.CODE & THE.AUTHOR.NAME & THE.DESCRIPTION \\'  # nosec B105
DEFAULT_REVISION = '00'
ROW_TEMPLATE_NAMED = r'issue & revision & author & summary \\'
ROW_TEMPLATE_ANONYMOUS = r'issue & revision & summary \\'
ROW_TEMPLATE_ANONYMOUS_VERSION = r'\centering issue & summary \\'
GLUE = '\n\\hline\n'
JSON_CHANNEL = 'json'
YAML_CHANNEL = 'yaml'
COLUMNS_EXPECTED = sorted(['author', 'date', 'issue', 'revision', 'summary'])
COLUMNS_MINIMAL = sorted(['issue', 'summary'])  # removed zero slot 'author' for data driven anonymous layout option
CUT_MARKER_CHANGES_TOP = '% |-- changes - cut - marker - top -->'
CUT_MARKER_CHANGES_BOTTOM = '% <-- changes - cut - marker - bottom --|'
CUT_MARKER_NOTICES_TOP = '% |-- notices - cut - marker - top -->'
CUT_MARKER_NOTICES_BOTTOM = '% <-- notices - cut - marker - bottom --|'
TOKEN_ADJUSTED_PUSHDOWN = r'\AdustedPushdown'  # nosec B105
DEFAULT_ADJUSTED_PUSHDOWN_VALUE = 14

LAYOUT_NAMED_CUT_MARKER_TOP = '% |-- layout named - cut - marker - top -->'
LAYOUT_NAMED_CUT_MARKER_BOTTOM = '% <-- layout named - cut - marker - bottom --|'
LAYOUT_ANONYMIZED_INSERT_MARKER = '% |-- layout anonymized - insert - marker --|'

NL = '\n'
TABLE_ANONYMOUS_PRE = r"""\
\begin{longtable}[]{|
  >{\raggedright\arraybackslash}p{(\columnwidth - 6\tabcolsep) * \real{0.0600}}|
  >{\raggedright\arraybackslash}p{(\columnwidth - 6\tabcolsep) * \real{0.0600}}|
  >{\raggedright\arraybackslash}p{(\columnwidth - 6\tabcolsep) * \real{0.8500}}|}
\hline
\begin{minipage}[b]{\linewidth}\raggedright
\textbf{\theChangeLogIssLabel}
\end{minipage} & \begin{minipage}[b]{\linewidth}\raggedright
\textbf{\theChangeLogRevLabel}
\end{minipage} & \begin{minipage}[b]{\linewidth}\raggedright
\textbf{\theChangeLogDescLabel}
\end{minipage} \\
\hline
"""
TABLE_ANONYMOUS_ROWS = r'THE.ISSUE.CODE & THE.REVISION.CODE & THE.DESCRIPTION \\'
TABLE_ANONYMOUS_POST = r"""\hline
\end{longtable}
"""

TABLE_ANONYMOUS_VERSION_PRE = r"""\
\begin{longtable}[]{|
  >{\raggedright\arraybackslash}p{(\columnwidth - 6\tabcolsep) * \real{0.0800}}|
  >{\raggedright\arraybackslash}p{(\columnwidth - 6\tabcolsep) * \real{0.8900}}|}
\hline
\begin{minipage}[b]{\linewidth}\raggedright
\textbf{\theChangeLogIssLabel}
\end{minipage} & \begin{minipage}[b]{\linewidth}\raggedright
\textbf{\hskip 12em \theChangeLogDescLabel}
\end{minipage} \\
\hline
"""
TABLE_ANONYMOUS_VERSION_ROWS = r'THE.ISSUE.CODE & THE.REVISION.CODE & THE.DESCRIPTION \\'
TABLE_ANONYMOUS_VERSION_POST = r"""\hline
\end{longtable}
"""


def get_layout(layout_path: PathLike, target_key: str, facet_key: str) -> dict[str, dict[str, dict[str, bool]]]:
    """Boolean layout decisions on bookmatter and publisher page conten.

    Deprecated as the known use cases evolved into a different direction ...
    """
    layout = {'layout': {'global': {'has_approvals': True, 'has_changes': True, 'has_notices': True}}}
    if layout_path:
        log.info(f'loading layout from {layout_path=} for changes and notices')
        return gat.load_layout(facet_key, target_key, layout_path)[0]  # type: ignore

    log.info('using default layout for approvals')
    return layout


def derive_model(model_path: PathLike) -> tuple[str, list[str]]:
    """Derive the model as channel type and column model from the given path."""
    channel = JSON_CHANNEL if str(model_path).endswith('.json') else YAML_CHANNEL
    columns_expected = COLUMNS_MINIMAL if channel == JSON_CHANNEL else COLUMNS_EXPECTED

    return channel, columns_expected


def columns_are_present(columns_present: list[str], columns_expected: list[str]) -> bool:
    """Ensure the needed columns are present."""
    return all(column in columns_expected for column in columns_present)


@no_type_check
def normalize(changes: object, channel: str, columns_expected: list[str]) -> list[dict[str, str]]:
    """Normalize the channel specific topology of the model into a logical model.

    On error an empty logical model is returned.
    """
    if channel == JSON_CHANNEL:
        for slot, change in enumerate(changes[0]['changes'], start=1):
            if not set(columns_expected).issubset(set(change)):
                log.error('unexpected column model!')
                log.error(f'-  expected: ({columns_expected})')
                log.error(f'-   minimal: ({COLUMNS_MINIMAL})')
                log.error(f'- but found: ({change}) for entry #{slot}')
                return []

    if channel == YAML_CHANNEL:
        for slot, change in enumerate(changes[0]['changes'], start=1):
            model = sorted(change.keys())
            if not set(COLUMNS_MINIMAL).issubset(set(model)):
                log.error('unexpected column model!')
                log.error(f'-  expected: ({columns_expected})')
                log.error(f'-   minimal: ({COLUMNS_MINIMAL})')
                log.error(f'- but found: ({model}) in slot {slot}')
                return []

    model = []
    if channel == JSON_CHANNEL:
        for change in changes[0]['changes']:
            issue, author, summary = change['issue'], change.get('author', None), change['summary']
            revision = change.get('revision', DEFAULT_REVISION)
            is_version = bool(change.get('version', False))
            model.append(
                {'issue': issue, 'revision': revision, 'author': author, 'summary': summary, 'is_version': is_version}
            )
        return model

    for change in changes[0]['changes']:
        author = change.get('author', None)
        issue = change['issue']
        revision = change.get('revision', DEFAULT_REVISION)
        is_version = bool(change.get('version', False))
        summary = change['summary']
        model.append(
            {'issue': issue, 'revision': revision, 'author': author, 'summary': summary, 'is_version': is_version}
        )

    return model


def adjust_pushdown_gen(text_lines: list[str], pushdown: float) -> Generator[str, None, None]:
    """Update the pushdown line filtering the incoming lines."""
    for line in text_lines:
        if TOKEN_ADJUSTED_PUSHDOWN in line:
            line = line.replace(TOKEN_ADJUSTED_PUSHDOWN, f'{pushdown}em')
            log.info(f'set adjusted pushdown value {pushdown}em')
        yield line


def weave(
    doc_root: Union[str, pathlib.Path],
    structure_name: str,
    target_key: str,
    facet_key: str,
    options: dict[str, Union[bool, str]],
    externals: ExternalsType,
) -> int:
    """Later alligator."""
    log.info(LOG_SEPARATOR)
    log.info('entered changes weave function ...')
    structure, asset_map = gat.prelude(
        doc_root=doc_root, structure_name=structure_name, target_key=target_key, facet_key=facet_key, command='changes'
    )

    layout_path = asset_map[target_key][facet_key].get(gat.KEY_LAYOUT, '')
    layout = get_layout(layout_path, target_key=target_key, facet_key=facet_key)
    log.info(f'{layout=}')

    log.info(LOG_SEPARATOR)
    changes_path = asset_map[target_key][facet_key][gat.KEY_CHANGES]
    channel, columns_expected = derive_model(changes_path)
    log.info(f'detected changes channel ({channel}) weaving in from ({changes_path})')

    log.info(f'loading changes from {changes_path=}')
    changes = gat.load_changes(facet_key, target_key, changes_path)
    log.info(f'{changes=}')

    log.info(LOG_SEPARATOR)
    log.info('plausibility tests for changes ...')

    logical_model = normalize(changes, channel=channel, columns_expected=columns_expected)

    is_anonymized = any(entry.get('author') is None for entry in logical_model)
    is_version_semantics = False
    if is_anonymized:
        if logical_model and logical_model[0].get('is_version', False):
            is_version_semantics = True
            rows = [
                ROW_TEMPLATE_ANONYMOUS_VERSION.replace('issue', kv['issue']).replace('summary', kv['summary'])
                for kv in logical_model
            ]
        else:
            rows = [
                ROW_TEMPLATE_ANONYMOUS.replace('issue', kv['issue'])
                .replace('revision', kv['revision'])
                .replace('summary', kv['summary'])
                for kv in logical_model
            ]
    else:
        rows = [
            ROW_TEMPLATE_NAMED.replace('issue', kv['issue'])
            .replace('revision', kv['revision'])
            .replace('author', kv['author'])
            .replace('summary', kv['summary'])
            for kv in logical_model
        ]

    pushdown = DEFAULT_ADJUSTED_PUSHDOWN_VALUE
    log.info(f'calculated adjusted pushdown to be {pushdown}em')

    publisher_template_is_custom = bool(externals['publisher']['is_custom'])
    publisher_template = str(externals['publisher']['id'])
    publisher_path = pathlib.Path('render/pdf/publisher.tex')

    publisher_template = tpl.load_resource(publisher_template, publisher_template_is_custom)
    lines = [line.rstrip() for line in publisher_template.split(NL)]

    if any(TOKEN_ADJUSTED_PUSHDOWN in line for line in lines):
        lines = list(adjust_pushdown_gen(lines, pushdown))
    else:
        log.error(f'token ({TOKEN_ADJUSTED_PUSHDOWN}) not found - template mismatch')

    if not layout['layout']['global']['has_changes']:
        log.info('removing changes from document layout')
        lines = list(too.remove_target_region_gen(lines, CUT_MARKER_CHANGES_TOP, CUT_MARKER_CHANGES_BOTTOM))
    elif is_anonymized:
        log.info('removing named changes table skeleton from document layout')
        lines = list(too.remove_target_region_gen(lines, LAYOUT_NAMED_CUT_MARKER_TOP, LAYOUT_NAMED_CUT_MARKER_BOTTOM))

    if not layout['layout']['global']['has_notices']:
        log.info('removing notices from document layout')
        lines = list(too.remove_target_region_gen(lines, CUT_MARKER_NOTICES_TOP, CUT_MARKER_NOTICES_BOTTOM))

    log.info(LOG_SEPARATOR)
    log.info('weaving in the changes from {changes_path} ...')
    for n, line in enumerate(lines):
        if is_anonymized:
            if line.strip() == LAYOUT_ANONYMIZED_INSERT_MARKER:
                if is_version_semantics:
                    table_text = TABLE_ANONYMOUS_VERSION_PRE + GLUE.join(rows) + TABLE_ANONYMOUS_VERSION_POST
                else:
                    table_text = TABLE_ANONYMOUS_PRE + GLUE.join(rows) + TABLE_ANONYMOUS_POST
                lines[n] = table_text
                break
        else:
            if line.strip() == CHANGE_ROW_TOKEN:
                lines[n] = GLUE.join(rows)
                break
    if lines[-1]:
        lines.append('\n')
    with open(publisher_path, 'wt', encoding=ENCODING) as handle:
        handle.write('\n'.join(lines))
    log.info(LOG_SEPARATOR)

    return 0



import os
import logging
import pathlib

from typer.testing import CliRunner

import liitos
import liitos.gather as gather
from liitos.cli import app

runner = CliRunner()

EXAMPLE_DEEP_PREFIX = pathlib.Path('example', 'deep')
ROLLUP_PATH = pathlib.Path('test', 'fixtures', 'rollup', '1')
TEST_PREFIX = pathlib.Path('test', 'fixtures', 'basic')
DEFAULT_STRUCTURE_PATH = TEST_PREFIX / gather.DEFAULT_STRUCTURE_NAME
TEST_TARGET = 'abc'
TEST_FACET = 'mn'
TEST_STRUCTURE = {
    TEST_TARGET: {
        TEST_FACET: {
            gather.KEY_APPROVALS: 'approvals.json',
            gather.KEY_BIND: 'bind-mn.txt',
            gather.KEY_CHANGES: 'changes.json',
            gather.KEY_META: 'meta-mn.md',
        }
    }
}
TEST_MAKE_MISSING = 'missing-this-file-for-'
OLE_WD = pathlib.Path.cwd()


def setup():
    print(pathlib.Path.cwd(), 'before')
    global OLE_WD
    OLE_WD = pathlib.Path.cwd()
    print(pathlib.Path.cwd(), 'during')


def teardown():
    os.chdir(OLE_WD)
    print(pathlib.Path.cwd(), 'after')


def test_version_ok():
    result = runner.invoke(app, ['version'])
    assert result.exit_code == 0
    assert f'version {liitos.__version__}' in result.stdout


def test_verify():
    result = runner.invoke(app, ['verify', '-f', 'mn', '-t', 'abc'])
    assert result.exit_code == 0


def test_verify_doc_root_option():
    result = runner.invoke(app, ['verify', '-d', f'{TEST_PREFIX}', '-f', 'mn', '-t', 'abc'])
    assert result.exit_code == 1  # TODO


def test_verify_pos():
    result = runner.invoke(app, ['verify', f'{TEST_PREFIX}', '-f', 'mn', '-t', 'abc'])
    assert result.exit_code == 0


def test_verify_pos_doc_root_not_present():
    result = runner.invoke(app, ['verify', f'{TEST_MAKE_MISSING}{TEST_PREFIX}', '-f', 'mn', '-t', 'abc'])
    assert result.exit_code == 0


def test_verify_pos_doc_root_no_folder():
    bad_location = f'{TEST_PREFIX}/{gather.DEFAULT_STRUCTURE_NAME}'
    result = runner.invoke(app, ['verify', bad_location, '-f', 'mn', '-t', 'abc'])
    assert result.exit_code == 0


def test_help():
    for options in ([], ['-h'], ['--help']):
        result = runner.invoke(app, options)
        assert result.exit_code == 0
        assert 'Verify the structure definition against the file system.' in result.stdout


def test_main():
    result = runner.invoke(app, ['verify', f'{TEST_PREFIX}', '-f', 'mn', '-t', 'abc'])
    assert result.exit_code == 0


def test_main_wrong_target():
    result = runner.invoke(app, ['verify', f'{TEST_PREFIX}', '-f', 'mn', '-t', 'no-target'])
    assert result.exit_code == 0


def test_main_wrong_facet():
    result = runner.invoke(app, ['verify', f'{TEST_PREFIX}', '-f', 'no-facet', '-t', 'abc'])
    assert result.exit_code == 0


def test_main_missing_asset():
    result = runner.invoke(app, ['verify', f'{TEST_PREFIX}', '-f', 'missing', '-t', 'abc'])
    assert result.exit_code == 0
    assert 'requested tree root at (test/fixtures/basic) does not exist' in result.stdout


def test_command_concat():
    result = runner.invoke(app, ['concat', f'{TEST_PREFIX}', '-f', 'mn', '-t', 'abc'])
    assert result.exit_code == 0


def test_command_render_base():
    result = runner.invoke(app, ['render', f'{TEST_PREFIX}', '-f', 'mn', '-t', 'abc'])
    assert result.exit_code == 2


def test_command_changes_base():
    result = runner.invoke(app, ['changes', f'{TEST_PREFIX}', '-f', 'mn', '-t', 'abc'])
    assert result.exit_code == 0


def test_command_approvals_base():
    result = runner.invoke(app, ['approvals', f'{TEST_PREFIX}', '-f', 'mn', '-t', 'abc'])
    assert result.exit_code == 0


def test_command_render_deep():
    os.chdir(OLE_WD)
    result = runner.invoke(
        app, ['render', '-d', f'{EXAMPLE_DEEP_PREFIX}', '-f', 'deep', '-s', 'structure.yml', '-t', 'prod_kind', '-v']
    )
    assert result.exit_code == 0


def test_command_render_approval_strategy_unknown():
    result = runner.invoke(
        app,
        [
            'render',
            '-d',
            f'{EXAMPLE_DEEP_PREFIX}',
            '-f',
            'deep',
            '-s',
            'structure.yml',
            '-t',
            'prod_kind',
            '-a',
            'no-layout',
        ],
    )
    assert result.exit_code == 2


def test_command_render_rollup(caplog):
    os.chdir(OLE_WD)
    caplog.set_level(logging.ERROR)
    result = runner.invoke(app, ['render', '-d', f'{ROLLUP_PATH}', '-f', 'rollup-1', '-t', 'prod_kind'])
    assert result.exit_code == 0
    assert not caplog.text


def test_command_render_rollup_eastwards(caplog):
    os.chdir(OLE_WD)
    caplog.set_level(logging.ERROR)
    result = runner.invoke(app, ['render', '-d', f'{ROLLUP_PATH}', '-f', 'rollup-1', '-t', 'prod_kind', '-a', 'east'])
    assert result.exit_code == 0
    assert not caplog.text


def test_command_report(caplog):
    caplog.set_level(logging.INFO)
    result = runner.invoke(app, ['report'])
    assert result.exit_code == 0
    assert 'tool-version-of-liitos process succeeded' in caplog.text


def test_command_reject_unknown(caplog):
    caplog.set_level(logging.ERROR)
    unknown_template = 'unknown-template'
    result = runner.invoke(app, ['eject', unknown_template])
    assert result.exit_code == 2
    assert f'eject of unknown template ({unknown_template}) requested' in caplog.text



"""Splice (Finnish liitos) contributions."""

import datetime as dti
import logging
import os
import pathlib
import shellingham  # type: ignore
from typing import Union, no_type_check

# [[[fill git_describe()]]]
__version__ = '2024.10.23+parent.g98d66860'
# [[[end]]] (checksum: c1c055967cc4e3beda66765b16a68c09)
__version_info__ = tuple(
    e if '-' not in e else e.split('-')[0] for part in __version__.split('+') for e in part.split('.') if e != 'parent'
)

APP_ALIAS = str(pathlib.Path(__file__).parent.name)
APP_ENV = APP_ALIAS.upper()
APP_NAME = locals()['__doc__']
DEBUG = bool(os.getenv(f'{APP_ENV}_DEBUG', ''))
VERBOSE = bool(os.getenv(f'{APP_ENV}_VERBOSE', ''))
QUIET = False
STRICT = bool(os.getenv(f'{APP_ENV}_STRICT', ''))
ENCODING = 'utf-8'
ENCODING_ERRORS_POLICY = 'ignore'
DEFAULT_CONFIG_NAME = f'.{APP_ALIAS}.json'

APP_VERSION = __version__
COMMA = ','
DEFAULT_LF_ONLY = 'YES'
FILTER_CS_LIST = 'mermaid-filter'
FROM_FORMAT_SPEC = 'markdown'
LATEX_PAYLOAD_NAME = 'document.tex'
log = logging.getLogger()  # Module level logger is sufficient
LOG_FOLDER = pathlib.Path('logs')
LOG_FILE = f'{APP_ALIAS}.log'
LOG_PATH = pathlib.Path(LOG_FOLDER, LOG_FILE) if LOG_FOLDER.is_dir() else pathlib.Path(LOG_FILE)
LOG_LEVEL = logging.INFO
LOG_SEPARATOR = '- ' * 80

DEFAULT_STRUCTURE_NAME = 'structure.yml'
KEY_APPROVALS = 'approvals'
KEY_BIND = 'bind'
KEY_CHANGES = 'changes'
KEY_LAYOUT = 'layout'
KEY_META = 'meta'
KEYS_REQUIRED = (KEY_APPROVALS, KEY_BIND, KEY_CHANGES, KEY_META)

CONTEXT: dict[str, str] = {}
KNOWN_APPROVALS_STRATEGIES = ('south', 'east')
APPROVALS_STRATEGY = os.getenv('LIITOS_APPROVALS_STRATEGY', '').lower()

PathLike = Union[str, pathlib.Path]
PathLikeOrBool = Union[PathLike, bool]
ExternalsType = dict[str, dict[str, PathLikeOrBool]]

try:
    SHELL = shellingham.detect_shell()
except shellingham.ShellDetectionFailure:
    SHELL = ('', 'echo')

TOOL_VERSION_COMMAND_MAP = {
    'etiketti': {
        'command': 'etiketti --version',
        'banner': 'Label and document the pdf file (data protection and identity)',
    },
    'exiftool': {
        'command': 'exiftool -ver',
        'banner': 'Change and list EXIF attributes',
    },
    'foran': {
        'command': 'foran version',
        'banner': 'Inspect local git status (or detect that there is no repository)',
    },
    'git': {
        'command': 'git --version',
        'banner': 'Version control system (git)',
    },
    'liitos': {
        'command': 'liitos version',
        'banner': 'Process the markdown documentation to produce PDF',
    },
    'lualatex': {
        'command': 'lualatex --version',
        'banner': 'Process LaTeX to produce PDF',
    },
    'mermaid': {
        'command': 'npm view mermaid',
        'banner': 'Mermaid for rendering diagrams from textual representations',
    },
    'mermaid-filter': {
        'command': 'npm view mermaid-filter',
        'banner': 'Pandoc filter for mermaid diagrams (rasterized version for PDF)',
    },
    'navigaattori': {
        'command': 'navigaattori version',
        'banner': 'Discover publication structural information from tree',
    },
    'node': {
        'command': 'node --version',
        'banner': 'Node server for executing some tools',
    },
    'npm': {
        'command': 'npm --version',
        'banner': 'Node package manager for inspecting versions of some node based tools',
    },
    'pandoc': {
        'command': 'pandoc --version',
        'banner': 'Pandoc for transforming markdown to LaTeX',
    },
    'pdfinfo': {
        'command': 'pdfinfo -v',
        'banner': 'Show PDF file information',
    },
    'python': {
        'command': 'python -V',
        'banner': 'Python driving it all',
    },
    'shell': {
        'command': f'{SHELL[1]} --version',
        'banner': 'The shell under which this process executes',
    },
    'svgexport': {
        'command': 'npm view svgexport',
        'banner': 'Export SVG to PNG (rasterized version for inclusion in PDF)',
    },
    'taksonomia': {
        'command': 'taksonomia --version',
        'banner': 'Assess and document the inventory of folders and files',
    },
}

ToolKey = str

EXTERNALS: ExternalsType = {
    'bookmatter': {
        'id': 'templates/bookmatter.tex.in',
        'is_custom': False,
    },
    'driver': {
        'id': 'templates/driver.tex.in',
        'is_custom': False,
    },
    'publisher': {
        'id': 'templates/publisher.tex.in',
        'is_custom': False,
    },
    'metadata': {
        'id': 'templates/metadata.tex.in',
        'is_custom': False,
    },
    'setup': {
        'id': 'templates/setup.tex.in',
        'is_custom': False,
    },
}

BOOKMATTER_TEMPLATE = os.getenv('LIITOS_BOOKMATTER_TEMPLATE', '')
if BOOKMATTER_TEMPLATE:
    EXTERNALS['bookmatter'] = {'id': BOOKMATTER_TEMPLATE, 'is_custom': True}

DRIVER_TEMPLATE = os.getenv('LIITOS_DRIVER_TEMPLATE', '')
if DRIVER_TEMPLATE:
    EXTERNALS['driver'] = {'id': DRIVER_TEMPLATE, 'is_custom': True}

METADATA_TEMPLATE = os.getenv('LIITOS_METADATA_TEMPLATE', '')
if METADATA_TEMPLATE:
    EXTERNALS['metadata'] = {'id': METADATA_TEMPLATE, 'is_custom': True}

PUBLISHER_TEMPLATE = os.getenv('LIITOS_PUBLISHER_TEMPLATE', '')
if PUBLISHER_TEMPLATE:
    EXTERNALS['publisher'] = {'id': PUBLISHER_TEMPLATE, 'is_custom': True}

SETUP_TEMPLATE = os.getenv('LIITOS_SETUP_TEMPLATE', '')
if SETUP_TEMPLATE:
    EXTERNALS['setup'] = {'id': SETUP_TEMPLATE, 'is_custom': True}

TS_FORMAT_LOG = '%Y-%m-%dT%H:%M:%S'
TS_FORMAT_PAYLOADS = '%Y-%m-%d %H:%M:%S.%f UTC'

__all__: list[str] = [
    'APP_ALIAS',
    'APP_ENV',
    'APP_VERSION',
    'APPROVALS_STRATEGY',
    'DEFAULT_STRUCTURE_NAME',
    'ENCODING',
    'EXTERNALS',
    'ExternalsType',
    'CONTEXT',
    'FILTER_CS_LIST',
    'FROM_FORMAT_SPEC',
    'KEY_APPROVALS',
    'KEY_BIND',
    'KEY_CHANGES',
    'KEY_LAYOUT',
    'KEY_META',
    'KEYS_REQUIRED',
    'KNOWN_APPROVALS_STRATEGIES',
    'LATEX_PAYLOAD_NAME',
    'LOG_SEPARATOR',
    'PathLike',
    'TOOL_VERSION_COMMAND_MAP',
    'ToolKey',
    'TS_FORMAT_PAYLOADS',
    'log',
    'parse_csl',
]


def parse_csl(csl: str) -> list[str]:
    """DRY."""
    return [fmt.strip() for fmt in csl.split(COMMA) if fmt.strip()]


@no_type_check
def formatTime_RFC3339(self, record, datefmt=None):  # noqa
    """HACK A DID ACK we could inject .astimezone() to localize ..."""
    return dti.datetime.fromtimestamp(record.created, dti.timezone.utc).isoformat()  # pragma: no cover


@no_type_check
def init_logger(name=None, level=None):
    """Initialize module level logger"""
    global log  # pylint: disable=global-statement

    log_format = {
        'format': '%(asctime)s %(levelname)s [%(name)s]: %(message)s',
        'datefmt': TS_FORMAT_LOG,
        # 'filename': LOG_PATH,
        'level': LOG_LEVEL if level is None else level,
    }
    logging.Formatter.formatTime = formatTime_RFC3339
    logging.basicConfig(**log_format)
    log = logging.getLogger(APP_ENV if name is None else name)
    log.propagate = True


init_logger(name=APP_ENV, level=logging.DEBUG if DEBUG else None)



import os
import pathlib

import liitos.concat as concat

BASIC_FIXTURE_ROOT = pathlib.Path('test', 'fixtures', 'basic')
EXAMPLE_DEEP_DOC_ROOT = pathlib.Path('example', 'deep')

RESTORE = os.getcwd()


def setup():
    os.chdir(RESTORE)


def teardown():
    os.chdir(RESTORE)


def test_adapt_image_images():
    collector = []
    caption = 'Cap...'
    img_path = 'x/images/abc.def'
    alt_text = '"Alt..."'
    text = f'![{caption}]({img_path} {alt_text})'
    assert concat.adapt_image(text, collector, 'x', root='y') == f'![{caption}](images/abc.def {alt_text})'
    assert collector == [f'{pathlib.Path().cwd()}/x/images/abc.def']


def test_adapt_image_diagrams():
    collector = []
    caption = 'Cap...'
    img_path = 'x/diagrams/abc.def'
    alt_text = '"Alt..."'
    text = f'![{caption}]({img_path} {alt_text})'
    assert concat.adapt_image(text, collector, 'x', root='y') == f'![{caption}](diagrams/abc.def {alt_text})'
    assert collector == [f'{pathlib.Path().cwd()}/x/diagrams/abc.def']


def test_adapt_image_other():
    collector = []
    # This may be not what anyone wants ...
    caption = 'Cap...'
    img_path = 'x/other/abc.def'
    alt_text = '"Alt..."'
    text = f'![{caption}]({img_path} {alt_text})'
    assert (
        concat.adapt_image(text, collector, 'x', root='y')
        == f'![{caption}]({pathlib.Path().cwd()}/x/other/abc.def {alt_text})'
    )
    assert collector == [f'{pathlib.Path().cwd()}/x/other/abc.def']


def test_adapt_image_dot_dot():
    collector = []
    # eg. example/deep/part/x.md has source ref to ../other/abc.def which is example/deep/other/abc.def
    caption = 'Cap...'
    img_path = '../other/abc.def'
    alt_text = '"Alt..."'
    text = f'![{caption}]({img_path} {alt_text})'
    assert (
        concat.adapt_image(text, collector, 'part/x.md', root='y')
        == f'![{caption}]({pathlib.Path().cwd()}/other/abc.def {alt_text})'
    )
    assert collector == [f'{pathlib.Path().cwd()}/other/abc.def']


def test_adapt_image_dot_dot_complete():
    """From example/deep/other/b.md."""
    collector = []
    caption = 'Caption Text Dot Dot Lime'
    img_path = '../images/lime.png'
    alt_text = '"Alt Text Dot Dot Lime"'
    text = f'![{caption}]({img_path} {alt_text})'
    assert concat.adapt_image(text, collector, 'other/b.md', root='y') == f'![{caption}](images/lime.png {alt_text})'
    assert collector == [f'{pathlib.Path().cwd()}/images/lime.png']


def test_parse_markdown_image():
    cases = {
        '! [Caption Text Red](images/red.png "Alt Text Red")': (
            '',
            '',
            '',
            '! [Caption Text Red](images/red.png "Alt Text Red")',
        ),
        ' ![Caption Text Red](images/red.png "Alt Text Red")': (
            '',
            '',
            '',
            ' ![Caption Text Red](images/red.png "Alt Text Red")',
        ),
        '![Caption Text Red] (images/red.png "Alt Text Red")': (
            '',
            '',
            '',
            '![Caption Text Red] (images/red.png "Alt Text Red")',
        ),
        '![ccc(sss "aaa") <!-- rest -->  ': (
            '',
            '',
            '',
            '![ccc(sss "aaa") <!-- rest -->  ',
        ),
        r'![cc\[c](sss "aaa") <!-- rest -->  ': (
            r'cc\[c',
            'sss',
            'aaa',
            ' <!-- rest -->  ',
        ),
        '![ccc](sss "aaa" <!-- rest -->  ': (
            '',
            '',
            '',
            '![ccc](sss "aaa" <!-- rest -->  ',
        ),
        '![ccc](sss)': (
            'ccc',
            'sss',
            '',
            '',
        ),
        '![cc(c](sss SSS "aaa") <!-- rest -->  ': (
            '',
            '',
            '',
            '![cc(c](sss SSS "aaa") <!-- rest -->  ',
        ),
        '![](sss "a(a)a")': (
            'INJECTED-CAP-TEXT-TO-MARK-MISSING-CAPTION-IN-OUTPUT',
            'sss',
            'a(a)a',
            '',
        ),
        '![captain](sss "a(a)a")': (
            'captain',
            'sss',
            'a(a)a',
            '',
        ),
        # This is not looking like anyone would want their alt text to be cut ...
        '![](sss "a(a)a" <!-- a remark you made -->': (
            'INJECTED-CAP-TEXT-TO-MARK-MISSING-CAPTION-IN-OUTPUT',
            'sss',
            'a(a',
            'a" <!-- a remark you made -->',
        ),
        # This is also not looking like anyone would want their alt text to be cut ...
        '![bla](sss "a(a)a"': (
            'bla',
            'sss',
            'a(a',
            'a"',
        ),
        '![Caption Text Red](images/red.png "Alt Text Red")': (
            'Caption Text Red',
            'images/red.png',
            'Alt Text Red',
            '',
        ),
        '![Caption Text Dot Dot Lime](../images/lime.png "Alt Text Dot Dot Lime")': (
            'Caption Text Dot Dot Lime',
            '../images/lime.png',
            'Alt Text Dot Dot Lime',
            '',
        ),
        '![Caption Text for SVG](/diagrams/squares-and-edges.svg) <!-- a comment -->  ': (
            'Caption Text for SVG',
            '/diagrams/squares-and-edges.svg',
            '',
            ' <!-- a comment -->  ',
        ),
        '![Caption Text for app specific SVG](diagrams/nuts-and-bolts.app.svg "Alt Text for app specific SVG")': (
            'Caption Text for app specific SVG',
            'diagrams/nuts-and-bolts.app.svg',
            'Alt Text for app specific SVG',
            '',
        ),
        r'![Caption \[Text] for app "specific SVG](diagrams/nuts-and-bolts.app.svg "Alt Text for ...")': (
            r'Caption \[Text] for app "specific SVG',
            'diagrams/nuts-and-bolts.app.svg',
            'Alt Text for ...',
            '',
        ),
        r'![Caption \[Text] for app "specific (SVG)](diagrams/nuts-and-bolts.app.svg "Alt Text ... SVG")': (
            r'Caption \[Text] for app "specific (SVG)',
            'diagrams/nuts-and-bolts.app.svg',
            'Alt Text ... SVG',
            '',
        ),
        '![](images/blue.png  "Alt Text Blue Same Repeated Image Caption Missing")': (
            'INJECTED-CAP-TEXT-TO-MARK-MISSING-CAPTION-IN-OUTPUT',
            'images/blue.png',
            'Alt Text Blue Same Repeated Image Caption Missing',
            '',
        ),
    }

    for text_line, expected in cases.items():
        assert concat.parse_markdown_image(text_line) == expected


def test_concatenate_base():
    parameters = {
        'doc_root': BASIC_FIXTURE_ROOT,
        'structure_name': 'structure.yml',
        'target_key': 'abc',
        'facet_key': 'mn',
        'options': {},
    }
    os.chdir(RESTORE)
    assert concat.concatenate(**parameters) == 0
    os.chdir(RESTORE)


def test_concatenate_deep():
    parameters = {
        'doc_root': EXAMPLE_DEEP_DOC_ROOT,
        'structure_name': 'structure.yml',
        'target_key': 'prod_kind',
        'facet_key': 'deep',
        'options': {},
    }
    os.chdir(RESTORE)
    assert concat.concatenate(**parameters) == 0
    os.chdir(RESTORE)



"""Loader function for placeholders (mostly images)."""

import pathlib
import pkgutil
from typing import Union

from liitos import ENCODING, PathLike, log

RESOURCES = (
    'placeholders/this-resource-is-missing.jpg',
    'placeholders/this-resource-is-missing.pdf',
    'placeholders/this-resource-is-missing.png',
    'placeholders/this-resource-is-missing.svg',
    'placeholders/this-resource-is-missing.tiff',
    'placeholders/this-resource-is-missing.webp',
)

READING_OPTIONS: dict[str, dict[str, Union[list[str], dict[str, str], None]]] = {
    '.jpg': {'args': ['rb'], 'kwargs': None},
    '.pdf': {'args': ['rb'], 'kwargs': None},
    '.png': {'args': ['rb'], 'kwargs': None},
    '.svg': {'args': ['rt'], 'kwargs': {'encoding': ENCODING}},
    '.tiff': {'args': ['rb'], 'kwargs': None},
    '.webp': {'args': ['rb'], 'kwargs': None},
}

WRITING_OPTIONS: dict[str, dict[str, Union[list[str], dict[str, str], None]]] = {
    '.jpg': {'args': ['wb'], 'kwargs': None},
    '.pdf': {'args': ['wb'], 'kwargs': None},
    '.png': {'args': ['wb'], 'kwargs': None},
    '.svg': {'args': ['wt'], 'kwargs': {'encoding': ENCODING}},
    '.tiff': {'args': ['wb'], 'kwargs': None},
    '.webp': {'args': ['wb'], 'kwargs': None},
}


def load_resource(resource: PathLike, is_complete_path: bool = False) -> tuple[str, Union[bytes, str]]:
    """Load the template either from the package resources or an external path."""
    from_path = pathlib.Path(resource)
    suffix = from_path.suffix
    if is_complete_path:
        if suffix and suffix in READING_OPTIONS:
            args = READING_OPTIONS[suffix].get('args')
            kwargs = READING_OPTIONS[suffix].get('kwargs')
            if READING_OPTIONS[suffix].get('kwargs'):
                with open(from_path, *args, **kwargs) as handle:  # type: ignore
                    return 'str', handle.read()
            with open(from_path, *args) as handle:  # type: ignore
                return 'bytes', handle.read()
        with open(from_path, 'rb') as handle:
            return 'bytes', handle.read()

    if suffix and suffix in READING_OPTIONS:
        args = READING_OPTIONS[suffix].get('args')
        kwargs = READING_OPTIONS[suffix].get('kwargs')
        if READING_OPTIONS[suffix].get('kwargs'):
            return 'str', pkgutil.get_data(__package__, str(resource)).decode(**kwargs)  # type: ignore
        return 'bytes', pkgutil.get_data(__package__, str(resource))  # type: ignore

    return 'bytes', pkgutil.get_data(__package__, str(resource))  # type: ignore


def eject(argv: Union[list[str], None] = None) -> int:
    """Eject the templates into the folder given (default MISSING) and create the folder if it does not exist."""
    argv = argv if argv else ['']
    into = argv[0]
    if not into.strip():
        into = 'MISSING'
    into_path = pathlib.Path(into)
    (into_path / 'placeholders').mkdir(parents=True, exist_ok=True)
    for resource in RESOURCES:
        write_to = into_path / resource
        suffix = write_to.suffix
        log.info(f'{resource} -> {write_to}')
        if suffix and suffix in WRITING_OPTIONS:
            args = WRITING_OPTIONS[suffix].get('args')
            kwargs = WRITING_OPTIONS[suffix].get('kwargs')
            if WRITING_OPTIONS[suffix].get('kwargs'):
                log.info(f'text({resource}) per ({args=}) and ({kwargs=})')
                data = pkgutil.get_data(__package__, resource).decode(**kwargs)  # type: ignore
                with open(write_to, *args, **kwargs) as target:  # type: ignore
                    target.write(data)
                continue
            log.info(f'binary({resource}) per ({args=})')
            data = pkgutil.get_data(__package__, resource)
            with open(write_to, *args) as target:  # type: ignore
                target.write(data)  # type: ignore
            continue
        log.warning(f'suffix ({suffix}) empty or not in ({", ".join(WRITING_OPTIONS.keys())})')

    return 0


def dump_placeholder(target: PathLike) -> tuple[int, str]:
    """Write out the placeholder matching the file type per suffix."""
    suffix = pathlib.Path(target).suffix
    proof = f'matching suffix ({suffix}) derived from target({target})'
    if suffix in WRITING_OPTIONS:
        args = WRITING_OPTIONS[suffix].get('args')
        kwargs = WRITING_OPTIONS[suffix].get('kwargs')
        resource = [res for res in RESOURCES if res.endswith(suffix)][0]
        kind, data = load_resource(resource)
        if kind == 'str':
            with open(target, *args, **kwargs) as handle:  # type: ignore
                handle.write(data)  # type: ignore
            return 0, f'wrote text mode placeholder {proof}'
        with open(target, *args) as handle:  # type: ignore
            handle.write(data)  # type: ignore
        return 0, f'wrote binary mode placeholder {proof}'

    return 1, f'no placeholder found {proof}'



import liitos.figures as figures


def test_scale_empty():
    assert figures.scale([]) == []


def test_scale_trigger_but_incomplete():
    assert figures.scale([r'\scale=']) == []  # This is may be surprising: error yes, but silently steal?


def test_scale_trigger_plus():
    incoming = [r'\scale=', r'\includegraphics{']
    assert figures.scale(incoming) == [
        r'\includegraphics{'
    ]  # This is may be surprising: error yes, but silently steal?


def test_scale_trigger_complete():
    incoming = ['', r'\scale=90%', '']
    assert figures.scale(incoming) == ['', '']  # This is may be surprising: error yes, but silently steal?


def test_scale_trigger_decimal_complete():
    incoming = ['', r'\scale=0.9', '']
    assert figures.scale(incoming) == ['', '']  # This is may be surprising: error yes, but silently steal?


def test_scale_trigger_complete_plus():
    incoming = ['', r'\scale=90%', '', r'\includegraphics{']
    assert figures.scale(incoming) == ['', '', r'\includegraphics{']


def test_scale_copy_through():
    incoming = ['', 'x', 'y', 'z']
    assert figures.scale(incoming) == incoming


def test_scale_trigger_percent_complete_plus():
    incoming = ['', r'\scale=90\%', '', 'text', r'\includegraphics{', '', 'x']
    outgoing = [
        '',
        '',
        'text',
        r'\includegraphics[width=0.9\textwidth,height=0.9\textheight,keepaspectratio]{',
        '',
        'x',
    ]
    assert figures.scale(incoming) == outgoing


def test_scale_trigger_percent_tiny__complete_plus():
    incoming = ['', r'\scale=0.90000000000000000001\%', '', 'text', r'\includegraphics{', '', 'x']
    outgoing = [
        '',
        '',
        'text',
        r'\includegraphics[width=0.01\textwidth,height=0.01\textheight,keepaspectratio]{',
        '',
        'x',
    ]
    assert figures.scale(incoming) == outgoing


def test_scale_trigger_fraction_complete_plus():
    incoming = ['', r'\scale=0.90000000000000000001', '', 'text', r'\includegraphics{', '', 'x']
    outgoing = [
        '',
        '',
        'text',
        r'\includegraphics[width=0.9\textwidth,height=0.9\textheight,keepaspectratio]{',
        '',
        'x',
    ]
    assert figures.scale(incoming) == outgoing


def test_scale_trigger_percent_complete_plus_another():
    incoming = [
        '',
        r'\scale=90\%',
        '',
        'text',
        r'\includegraphics{',
        '',
        'x',
        '',
        r'\scale=92\%',
        '',
        'texte',
        r'\includegraphics{x',
        '',
        'x',
    ]
    outgoing = [
        '',
        '',
        'text',
        r'\includegraphics[width=0.9\textwidth,height=0.9\textheight,keepaspectratio]{',
        '',
        'x',
        '',
        '',
        'texte',
        r'\includegraphics[width=0.92\textwidth,height=0.92\textheight,keepaspectratio]{x',
        '',
        'x',
    ]
    assert figures.scale(incoming) == outgoing



import os
import pathlib

import liitos.render as render
from liitos import FILTER_CS_LIST

BASIC_FIXTURE_ROOT = pathlib.Path('test', 'fixtures', 'basic')
EXAMPLE_DEEP_DOC_ROOT = pathlib.Path('example', 'deep')

RESTORE = os.getcwd()


def setup():
    os.chdir(RESTORE)


def teardown():
    os.chdir(RESTORE)


def test_ren_der():
    parameters = {
        'doc_root': EXAMPLE_DEEP_DOC_ROOT,
        'structure_name': 'structure.yml',
        'target_key': 'prod_kind',
        'facet_key': 'deep',
        'options': {
            'filter_cs_list': FILTER_CS_LIST,
        },
    }
    os.chdir(RESTORE)
    assert render.der(**parameters) == 0
    os.chdir(RESTORE)



import os
import pathlib

import liitos.approvals as approvals
from liitos import EXTERNALS

BASIC_FIXTURE_ROOT = pathlib.Path('test', 'fixtures', 'basic')
EXAMPLE_DEEP_DOC_ROOT = pathlib.Path('example', 'deep')
EXAMPLE_LEGACY_DOC_ROOT = pathlib.Path('example', 'legacy')
LAYOUT_ALL_PATH = pathlib.Path('test/fixtures/layout/all.yml')


def test_approvals():
    parameters = {
        'doc_root': EXAMPLE_DEEP_DOC_ROOT,
        'structure_name': 'structure.yml',
        'target_key': 'prod_kind',
        'facet_key': 'deep',
        'options': {},
        'externals': EXTERNALS,
    }
    restore = os.getcwd()
    assert approvals.weave(**parameters) == 0
    os.chdir(restore)


def test_approvals_legacy():
    parameters = {
        'doc_root': EXAMPLE_LEGACY_DOC_ROOT,
        'structure_name': 'structure.yml',
        'target_key': 'prod_kind',
        'facet_key': 'legacy',
        'options': {},
        'externals': EXTERNALS,
    }
    restore = os.getcwd()
    assert approvals.weave(**parameters) == 0
    os.chdir(restore)


def test_normalize_json_columns_mismatch():
    signatures = [{'columns': ['columns', 'are', 'unexpected']}]
    channel = approvals.JSON_CHANNEL
    columns_expected = approvals.COLUMNS_EXPECTED
    assert approvals.normalize(signatures=signatures, channel=channel, columns_expected=columns_expected) == []


def test_normalize_yaml_columns_mismatch():
    signatures = [{'approvals': [{'columns': 'are', 'not': 'expected'}]}]
    channel = approvals.YAML_CHANNEL
    columns_expected = approvals.COLUMNS_EXPECTED
    assert approvals.normalize(signatures=signatures, channel=channel, columns_expected=columns_expected) == []


def test_eastern_scaffold():
    normalized = [{'role': 'role', 'name': 'name'}]
    table = approvals.eastern_scaffold(normalized)
    assert table
    assert 'THE.ORGA0.SLOT' in table
    assert 'THE.ROLE0.SLOT' in table
    assert 'THE.NAME0.SLOT' in table


def test_inject_eastwards():
    lines = ['a']
    expected = [
        'a',
        '% |-- layout east - cut - marker - top -->',
        r'\begin{small}',
        r'\addtolength\aboverulesep{0.15ex}  % extra spacing above and below rules',
        r'\addtolength\belowrulesep{0.35ex}',
        r'\begin{longtable}[]{|',
        r' >{\raggedright\arraybackslash}m{(\columnwidth - 12\tabcolsep) * \real{0.2000}}|% <- fixed',
        r'}',
        r'\hline',
        r'\begin{minipage}[b]{\linewidth}\raggedright\ \centering \textbf{\theApprovalsDepartmentLabel}\end{minipage}%',
        r' \\[0.5ex]',
        r'\hline',
        r'\ \mbox{\textbf{\theApprovalsRoleLabel}}%',
        '',
        r' \\[0.5ex]',
        r'\hline',
        r'\ \mbox{\textbf{\theApprovalsNameLabel}}%',
        '',
        r' \\[0.5ex]',
        r'\hline',
        r'\ \mbox{\textbf{Date}} \mbox{\textbf{\ \ \ \ \ \ }} \mbox{\textbf{\ Signature}}%',
        r'',
        r' \\[0.5ex]',
        r'\hline',
        '',
        r'\end{longtable}',
        r'\end{small}',
        '% <-- layout east - cut - marker - bottom --|',
        '',
    ]
    approvals.inject_eastwards(lines, normalized=[], pushdown=42.0)
    assert lines == expected


def test_inject_eastwards_many():
    lines = ['a']
    normalized = [
        {'orga': f'orga{n}', 'role': f'rolen{n}', 'name': f'name{n}'}
        for n in range(1, approvals.EASTERN_TABLE_MAX_MEMBERS + 2)
    ]
    expected = [
        'a',
        '% |-- layout east - cut - marker - top -->',
        r'\begin{small}',
        r'\addtolength\aboverulesep{0.15ex}  % extra spacing above and below rules',
        r'\addtolength\belowrulesep{0.35ex}',
        r'\begin{longtable}[]{|',
        r' >{\raggedright\arraybackslash}m{(\columnwidth - 12\tabcolsep) * \real{0.2000}}|% <- fixed',
        r' >{\raggedright\arraybackslash}m{(\columnwidth - 12\tabcolsep) * \real{0.2000}}|',
        r' >{\raggedright\arraybackslash}m{(\columnwidth - 12\tabcolsep) * \real{0.2000}}|',
        r' >{\raggedright\arraybackslash}m{(\columnwidth - 12\tabcolsep) * \real{0.2000}}|',
        r' >{\raggedright\arraybackslash}m{(\columnwidth - 12\tabcolsep) * \real{0.2000}}|}',
        r'\hline',
        r'\begin{minipage}[b]{\linewidth}\raggedright\ '
        r'\centering \textbf{\theApprovalsDepartmentLabel}\end{minipage}%',
        r' & \begin{minipage}[b]{\linewidth}\centering\arraybackslash \textbf{orga1}\end{minipage}',
        r' & \begin{minipage}[b]{\linewidth}\centering\arraybackslash \textbf{orga2}\end{minipage}',
        r' & \begin{minipage}[b]{\linewidth}\centering\arraybackslash \textbf{orga3}\end{minipage}',
        r' & \begin{minipage}[b]{\linewidth}\centering\arraybackslash \textbf{orga4}\end{minipage} \\[0.5ex]',
        r'\hline',
        r'\ \mbox{\textbf{\theApprovalsRoleLabel}}%',
        r' & \centering\arraybackslash \textbf{rolen1}',
        r' & \centering\arraybackslash \textbf{rolen2}',
        r' & \centering\arraybackslash \textbf{rolen3}',
        r' & \centering\arraybackslash \textbf{rolen4}',
        r' \\[0.5ex]',
        r'\hline',
        r'\ \mbox{\textbf{\theApprovalsNameLabel}}%',
        r' & \centering\arraybackslash name1',
        r' & \centering\arraybackslash name2',
        r' & \centering\arraybackslash name3',
        r' & \centering\arraybackslash name4',
        r' \\[0.5ex]',
        r'\hline',
        r'\ \mbox{\textbf{Date}} \mbox{\textbf{\ \ \ \ \ \ }} \mbox{\textbf{\ Signature}}%',
        r' & \mbox{}',
        r' & \mbox{}',
        r' & \mbox{}',
        r' & \mbox{}',
        '',
        r' \\[0.5ex]',
        r'\hline',
        '',
        r'\end{longtable}',
        r'\end{small}',
        '% <-- layout east - cut - marker - bottom --|',
        '',
        '% |-- layout east - cut - marker - top -->',
        r'\begin{small}',
        r'\addtolength\aboverulesep{0.15ex}  % extra spacing above and below rules',
        r'\addtolength\belowrulesep{0.35ex}',
        r'\begin{longtable}[]{|',
        r' >{\raggedright\arraybackslash}m{(\columnwidth - 12\tabcolsep) * \real{0.2000}}|% <- fixed',
        r' >{\raggedright\arraybackslash}m{(\columnwidth - 12\tabcolsep) * \real{0.2000}}|}',
        r'\hline',
        r'\begin{minipage}[b]{\linewidth}\raggedright\ '
        r'\centering \textbf{\theApprovalsDepartmentLabel}\end{minipage}%',
        r' & \begin{minipage}[b]{\linewidth}\centering\arraybackslash \textbf{orga5}\end{minipage} \\[0.5ex]',
        r'\hline',
        r'\ \mbox{\textbf{\theApprovalsRoleLabel}}%',
        r' & \centering\arraybackslash \textbf{rolen5}',
        r' \\[0.5ex]',
        r'\hline',
        r'\ \mbox{\textbf{\theApprovalsNameLabel}}%',
        r' & \centering\arraybackslash name5',
        r' \\[0.5ex]',
        r'\hline',
        r'\ \mbox{\textbf{Date}} \mbox{\textbf{\ \ \ \ \ \ }} \mbox{\textbf{\ Signature}}%',
        r' & \mbox{}',
        '',
        r' \\[0.5ex]',
        r'\hline',
        '',
        r'\end{longtable}',
        r'\end{small}',
        '% <-- layout east - cut - marker - bottom --|',
        '',
    ]
    approvals.inject_eastwards(lines, normalized=normalized, pushdown=3.14156)
    assert lines == expected


def test_inject_southwards():
    lines = ['a']
    approvals.inject_southwards(lines, rows=[], pushdown=42.0)
    assert lines == ['a', '\n']


def test_get_layout_from_path():
    data = approvals.get_layout(LAYOUT_ALL_PATH, 'target', 'facet')
    assert data == {'layout': {'global': {'has_approvals': True, 'has_changes': True, 'has_notices': True}}}



import os
import pathlib

import liitos.changes as changes
from liitos import EXTERNALS

EXAMPLE_DEEP_DOC_ROOT = pathlib.Path('example', 'deep')
EXAMPLE_LEGACY_DOC_ROOT = pathlib.Path('example', 'legacy')
LAYOUT_ALL_PATH = pathlib.Path('test/fixtures/layout/all.yml')


def test_changes_deep():
    parameters = {
        'doc_root': EXAMPLE_DEEP_DOC_ROOT,
        'structure_name': 'structure.yml',
        'target_key': 'prod_kind',
        'facet_key': 'deep',
        'options': {},
        'externals': EXTERNALS,
    }
    restore = os.getcwd()
    assert changes.weave(**parameters) == 0
    os.chdir(restore)


def test_changes_legacy():
    parameters = {
        'doc_root': EXAMPLE_LEGACY_DOC_ROOT,
        'structure_name': 'structure.yml',
        'target_key': 'prod_kind',
        'facet_key': 'legacy',
        'options': {},
        'externals': EXTERNALS,
    }
    restore = os.getcwd()
    assert changes.weave(**parameters) == 0
    os.chdir(restore)


def test_normalize_json_columns_mismatch():
    changes_in = [{'changes': ['columns', 'are', 'unexpected']}]
    channel = changes.JSON_CHANNEL
    columns_expected = changes.COLUMNS_EXPECTED
    assert changes.normalize(changes=changes_in, channel=channel, columns_expected=columns_expected) == []


def test_normalize_yaml_columns_mismatch():
    changes_in = [{'changes': [{'columns': 'are', 'not': 'expected'}]}]
    channel = changes.YAML_CHANNEL
    columns_expected = changes.COLUMNS_EXPECTED
    assert changes.normalize(changes=changes_in, channel=channel, columns_expected=columns_expected) == []


def test_get_layout_from_path():
    data = changes.get_layout(LAYOUT_ALL_PATH, 'target', 'facet')
    assert data == {'layout': {'global': {'has_approvals': True, 'has_changes': True, 'has_notices': True}}}
